from __future__ import annotations

import importlib
import importlib.metadata
from dataclasses import dataclass

from loguru import logger
from omu.result import Err, Ok, Result

from .loader import normalize_package_name

PLUGIN_GROUP = "omu.plugins"


@dataclass(frozen=True, slots=True)
class PluginEntry:
    key: str
    name: str
    entry_point: importlib.metadata.EntryPoint
    dist: importlib.metadata.Distribution

    @classmethod
    def from_entrypoint(cls, entry_point: importlib.metadata.EntryPoint) -> Result[PluginEntry, str]:
        assert entry_point.dist is not None
        if entry_point.dist is None:
            return Err(f"Invalid plugin: {entry_point} has no distribution")
        package = entry_point.dist.name
        return Ok(
            PluginEntry(
                key=entry_point.value,
                name=normalize_package_name(package),
                entry_point=entry_point,
                dist=entry_point.dist,
            )
        )

    @classmethod
    def retrieve_from_distribution(
        cls,
        distribution: importlib.metadata.Distribution,
    ) -> dict[str, PluginEntry]:
        entries: dict[str, PluginEntry] = {}
        entry_points = distribution.entry_points
        for entry_point in entry_points:
            if entry_point.group != PLUGIN_GROUP:
                continue
            match PluginEntry.from_entrypoint(entry_point):
                case Ok(entry):
                    entries[entry.key] = entry
                case Err(message):
                    logger.warning(f"Invalid plugin {entry_point}: {message}")
        return entries

    @classmethod
    def retrieve_plugin_entries(cls) -> dict[str, PluginEntry]:
        plugin_entries: dict[str, PluginEntry] = {}
        entry_points = importlib.metadata.entry_points(group=PLUGIN_GROUP)
        for entry_point in entry_points:
            match PluginEntry.from_entrypoint(entry_point):
                case Ok(entry):
                    if entry.name in plugin_entries:
                        logger.warning(f"Duplicate plugin: {entry_point}")
                        continue
                    plugin_entries[entry.name] = entry
                case Err(message):
                    logger.warning(f"Invalid plugin {entry_point}: {message}")
        return plugin_entries



from collections.abc import AsyncGenerator, Mapping

from omu.api.table import Table, TableConfig, TableEvents, TablePermissions, TableType
from omu.event_emitter import Unlisten
from omu.helper import AsyncCallback, Coro
from omu.identifier import Identifier
from omu.serializer import Serializable

from .server_table import ServerTable


class SerializeAdapter[T](dict[str, T]):
    def __init__(self, cache: Mapping[str, bytes], serializer: Serializable[T, bytes]):
        self._cache = cache
        self._serializer = serializer

    def __getitem__(self, key: str) -> T:
        return self._serializer.deserialize(self._cache[key])


class SerializedTable[T](Table[T]):
    def __init__(self, table: ServerTable, type: TableType[T]):
        self._table = table
        self._type = type
        self._event = TableEvents[T](self)
        self._proxies: list[Coro[[T], T | None]] = []
        self._chunk_size = 100
        self._permissions: TablePermissions | None = None
        self.permission_read: Identifier | None = None
        self.permission_write: Identifier | None = None
        self._listening = False
        table.event.cache_update += self.on_cache_update
        table.event.add += self.on_add
        table.event.update += self.on_update
        table.event.remove += self.on_remove
        table.event.clear += self.on_clear

    @property
    def cache(self) -> Mapping[str, T]:
        return SerializeAdapter(self._table.cache, self._type.serializer)

    def set_permissions(
        self,
        /,
        all: Identifier | None = None,
        read: Identifier | None = None,
        write: Identifier | None = None,
        remove: Identifier | None = None,
        proxy: Identifier | None = None,
    ) -> None:
        self._permissions = TablePermissions(
            all=all,
            read=read,
            write=write,
            remove=remove,
            proxy=proxy,
        )
        self._table.set_permissions(self._permissions)

    def set_config(self, config: TableConfig) -> None:
        self._table.set_config(config)

    def set_cache_size(self, size: int) -> None:
        self._table.set_cache_size(size)

    async def get(self, key: str) -> T | None:
        if key in self._table.cache:
            return self._type.serializer.deserialize(self._table.cache[key])
        item = await self._table.get(key)
        if item is None:
            return None
        return self._type.serializer.deserialize(item)

    async def get_many(self, *keys: str) -> dict[str, T]:
        items = await self._table.get_many(*keys)
        return {key: self._type.serializer.deserialize(item) for key, item in items.items()}

    async def add(self, *items: T) -> None:
        data = {self._type.key_function(item): self._type.serializer.serialize(item) for item in items}
        await self._table.add(data)

    async def update(self, *items: T) -> None:
        data = {self._type.key_function(item): self._type.serializer.serialize(item) for item in items}
        await self._table.update(data)

    async def remove(self, *items: T) -> None:
        await self._table.remove([self._type.key_function(item) for item in items])

    async def clear(self) -> None:
        await self._table.clear()

    async def has(self, key: str) -> bool:
        return await self._table.has(key)

    async def has_many(self, *keys: str) -> dict[str, bool]:
        return await self._table.has_many(list(keys))

    async def has_all(self, *keys: str) -> bool:
        return await self._table.has_all(list(keys))

    async def has_any(self, *keys: str) -> bool:
        return await self._table.has_any(list(keys))

    async def fetch_items(
        self,
        limit: int,
        backward: bool = False,
        cursor: str | None = None,
    ) -> dict[str, T]:
        items = await self._table.fetch_items(limit, backward, cursor)
        return self._parse_items(items)

    async def fetch_range(self, start: str, end: str) -> dict[str, T]:
        items = await self._table.fetch_range(start, end)
        return self._parse_items(items)

    async def fetch_all(self) -> dict[str, T]:
        items = await self._table.fetch_all()
        return self._parse_items(items)

    async def iterate(
        self,
        backward: bool = False,
        cursor: str | None = None,
    ) -> AsyncGenerator[T, None]:
        while True:
            items = await self.fetch_items(
                limit=self._chunk_size,
                backward=backward,
                cursor=cursor,
            )
            if len(items) == 0:
                break
            for item in items.values():
                yield item
            cursor = next(iter(items.keys()))

    async def size(self) -> int:
        return await self._table.size()

    @property
    def event(self) -> TableEvents[T]:
        return self._event

    def listen(self, listener: AsyncCallback[Mapping[str, T]] | None = None) -> Unlisten:
        self._listening = True
        if listener:
            return self._event.cache_update.listen(listener)
        return lambda: None

    async def on_cache_update(self, cache: Mapping[str, bytes]) -> None:
        await self._event.cache_update(self._parse_items(cache))

    async def on_add(self, items: Mapping[str, bytes]) -> None:
        _items = self._parse_items(items)
        await self._event.add(_items)

    async def on_update(self, items: Mapping[str, bytes]) -> None:
        _items = self._parse_items(items)
        await self._event.update(_items)

    async def on_remove(self, items: Mapping[str, bytes]) -> None:
        _items = self._parse_items(items)
        await self._event.remove(_items)

    async def on_clear(self) -> None:
        await self._event.clear()

    def proxy(self, callback: Coro[[T], T | None]) -> Unlisten:
        raise NotImplementedError

    def _parse_items(self, items: Mapping[str, bytes]) -> dict[str, T]:
        parsed: dict[str, T] = {}
        for key, item in items.items():
            item = self._type.serializer.deserialize(item)
            if not item:
                raise Exception(f"Failed to deserialize item {key}")
            parsed[key] = item
        return parsed



from __future__ import annotations

import os
from collections.abc import Callable, Mapping
from typing import TYPE_CHECKING

from omu.api.server.extension import (
    SERVER_APP_TABLE_TYPE,
    SERVER_INDEX_REGISTRY_TYPE,
    SHUTDOWN_ENDPOINT_TYPE,
    TRUSTED_HOSTS_REGISTRY_TYPE,
)
from omu.app import App, AppType
from omu.network.packet.packet_types import DisconnectType

from omuserver.session import Session

from .permissions import (
    SERVER_APPS_READ_PERMISSION,
    SERVER_APPS_WRITE_PERMISSION,
    SERVER_INDEX_READ_PERMISSION,
    SERVER_SHUTDOWN_PERMISSION,
    SERVER_TRUSTED_ORIGINS_GET_PERMISSION,
)

if TYPE_CHECKING:
    from loguru import Message

    from omuserver.server import Server


class LogHandler:
    def __init__(
        self,
        callback: Callable[[str], None],
    ) -> None:
        self.callback = callback

    def write(self, message: Message) -> None:
        self.callback(message)


class ServerExtension:
    def __init__(self, server: Server) -> None:
        self._server = server
        server.security.register_permission(
            SERVER_SHUTDOWN_PERMISSION,
            SERVER_INDEX_READ_PERMISSION,
            SERVER_APPS_READ_PERMISSION,
            SERVER_APPS_WRITE_PERMISSION,
            SERVER_TRUSTED_ORIGINS_GET_PERMISSION,
        )
        server.endpoints.bind(SHUTDOWN_ENDPOINT_TYPE, self.handle_shutdown)
        self.index = self._server.registries.register(SERVER_INDEX_REGISTRY_TYPE)
        self.apps = self._server.tables.register(SERVER_APP_TABLE_TYPE)
        self.apps.event.remove += self.on_remove_app
        self.apps.event.add += self.cleanup
        self.trusted_hosts = self._server.registries.register(TRUSTED_HOSTS_REGISTRY_TYPE)

    async def on_remove_app(self, items: Mapping[str, App]):
        for app in items.values():
            self._server.security.remove_app(app.id)
            session = self._server.sessions.sessions.get(app.id)
            if session:
                await session.disconnect(DisconnectType.APP_REMOVED, f"App {app.id.key()} removed")
        await self.cleanup()

    async def cleanup(self, *_):
        apps = await self.apps.fetch_all()
        unused_services = {k: app for k, app in apps.items() if app.type == AppType.SERVICE}
        for app in apps.values():
            if app.dependencies is None:
                continue

            for dependency in app.dependencies.keys():
                if dependency not in unused_services:
                    continue

                del unused_services[dependency]

        unparented_apps: set[App] = set()
        for app in apps.values():
            if app.parent_id is None:
                continue
            found_app = apps.get(app.parent_id.key())
            if found_app is None:
                unparented_apps.add(app)

        for app in apps.values():
            if app.dependencies is None:
                continue

            for dependency in app.dependencies.keys():
                if dependency not in unused_services:
                    continue

                del unused_services[dependency]

        if len(unused_services) == 0 and len(unparented_apps) == 0:
            return
        to_remove = [*unparented_apps, *unused_services.values()]
        for app in to_remove:
            self._server.security.remove_app(app.id)
        await self.apps.remove(*to_remove)

    async def handle_shutdown(self, session: Session, restart: bool = False) -> bool:
        await self.shutdown(restart)
        return True

    async def shutdown(self, restart: bool = False) -> None:
        try:
            if restart:
                await self._server.restart()
            else:
                await self._server.stop()
        finally:
            os._exit(0)



from omu.api.http.extension import HTTP_REQUEST_PERMISSION_ID
from omu.api.permission.permission import PermissionType

HTTP_REQUEST_PERMISSION = PermissionType(
    HTTP_REQUEST_PERMISSION_ID,
    {
        "level": "low",
        "name": {
            "en": "HTTP Requests",
            "ja": "HTTPリクエスト",
        },
        "note": {
            "ja": "アプリが外部のサイトにアクセスすることを許可します。",
            "en": "Allows apps to access external sites.",
        },
    },
)



from __future__ import annotations

from typing import TYPE_CHECKING

from omu.api.i18n.extension import I18N_LOCALES_REGISTRY_TYPE

from .permissions import I18N_GET_LOCALES_PERMISSION, I18N_SET_LOCALES_PERMISSION

if TYPE_CHECKING:
    from omuserver.server import Server


class I18nExtension:
    def __init__(self, server: Server):
        server.security.register_permission(
            I18N_GET_LOCALES_PERMISSION,
            I18N_SET_LOCALES_PERMISSION,
        )
        server.registries.register(I18N_LOCALES_REGISTRY_TYPE)



from __future__ import annotations

import io
import ipaddress
import socket
from asyncio import Future
from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import aiohttp
from aiohttp import ClientResponse, ClientSession, ClientWebSocketResponse, WSCloseCode, WSMsgType
from omu.api.http.extension import (
    HTTP_ALLOWED_PORTS,
    HTTP_REQUEST_CLOSE,
    HTTP_REQUEST_CREATE,
    HTTP_REQUEST_PERMISSION_ID,
    HTTP_REQUEST_SEND,
    HTTP_RESPONSE_CHUNK,
    HTTP_RESPONSE_CLOSE,
    HTTP_RESPONSE_CREATE,
    WEBSOCKET_CLOSE,
    WEBSOCKET_CREATE,
    WEBSOCKET_DATA,
    WEBSOCKET_ERROR,
    WEBSOCKET_OPEN,
    DataChunk,
    HttpRequest,
    HttpResponse,
    RequestHandle,
    WebSocketClose,
    WSDataMeta,
)
from omu.app import App, AppType
from omu.errors import PermissionDenied
from omu.identifier import Identifier
from yarl import URL

from omuserver.api.http.permission import HTTP_REQUEST_PERMISSION
from omuserver.consts import USER_AGENT_HEADERS
from omuserver.session import Session

if TYPE_CHECKING:
    from omuserver.server import Server


def serialize_response(id: str, response: ClientResponse) -> HttpResponse:
    return {
        "id": id,
        "header": dict(response.headers),
        "history": [serialize_response(id, resp) for resp in response.history],
        "redirected": len(response.history) > 0,
        "status": response.status,
        "statusText": response.reason,
        "url": str(response.url),
    }


@dataclass(frozen=True, slots=True)
class HttpHandle:
    session: Session
    buffer: io.BytesIO
    close_future: Future[io.BytesIO] = field(default_factory=Future)


@dataclass(slots=True)
class WSHandle:
    session: Session
    socket: Future[aiohttp.ClientWebSocketResponse]


class NotAllowed(Exception): ...


class HttpExtension:
    def __init__(self, server: Server):
        self.server = server
        self.http_handles: dict[str, HttpHandle] = {}
        self.ws_handles: dict[str, WSHandle] = {}
        server.security.register_permission(HTTP_REQUEST_PERMISSION)
        server.network.register_packet(
            HTTP_REQUEST_CREATE,
            HTTP_REQUEST_SEND,
            HTTP_REQUEST_CLOSE,
            WEBSOCKET_CREATE,
            WEBSOCKET_OPEN,
            WEBSOCKET_CLOSE,
            WEBSOCKET_DATA,
            WEBSOCKET_ERROR,
        )
        server.network.add_packet_handler(HTTP_REQUEST_CREATE, self.handle_http_request_create)
        server.network.add_packet_handler(HTTP_REQUEST_SEND, self.handle_http_request_send)
        server.network.add_packet_handler(HTTP_REQUEST_CLOSE, self.handle_http_request_close)
        server.network.add_packet_handler(WEBSOCKET_CREATE, self.handle_ws_create)
        server.network.add_packet_handler(WEBSOCKET_DATA, self.handle_ws_send)
        server.network.add_packet_handler(WEBSOCKET_CLOSE, self.handle_ws_close)
        self.allowed_ports = server.tables.register(HTTP_ALLOWED_PORTS)
        server.server.apps.event.remove += self.on_app_removed

    async def on_app_removed(self, apps: Mapping[str, App]):
        for id in apps.keys():
            found = await self.allowed_ports.get(id)
            if found is None:
                continue
            await self.allowed_ports.remove(found)

    async def verify_port_allowed(self, session: Session, url: URL):
        if session.kind == AppType.DASHBOARD:
            return
        if url.host is None:
            raise Exception("Invalid url: host is not specified")
        ip_address = socket.gethostbyname(url.host)
        address = ipaddress.ip_address(ip_address)
        if address.is_global:
            return
        if url.port is None:
            return
        id = session.id.key()
        entry = await self.allowed_ports.get(id) or {"id": id, "ports": []}
        if url.port in entry["ports"]:
            return
        accepted = await self.server.dashboard.notify_http_port_request(session.app, url.port)
        if not accepted:
            raise NotAllowed(f"App {id} does not have permission to access port {url.port}.")
        entry["ports"].append(url.port)
        await self.allowed_ports.update(entry)

    async def handle_http_request_create(self, session: Session, packet: HttpRequest):
        if not session.permissions.has(HTTP_REQUEST_PERMISSION_ID):
            raise PermissionDenied(f"Missing HTTP request permission: {HTTP_REQUEST_PERMISSION_ID}")
        id = Identifier.from_key(packet["id"])
        if not session.is_app_id(id):
            raise Exception(f"Request id {packet['id']} is not a app id of {session.id.key()}")
        if packet["id"] in self.http_handles:
            raise Exception(f"Request id {packet['id']} already existing")
        try:
            url = URL(packet["url"])
            request = HttpHandle(session=session, buffer=io.BytesIO())
            await self.verify_port_allowed(session, url)
            self.http_handles[packet["id"]] = request
            await request.close_future
            async with ClientSession(headers=USER_AGENT_HEADERS) as client:
                async with client.request(
                    packet["method"],
                    url,
                    headers=packet["header"],
                    allow_redirects=packet["redirect"] == "follow",
                    data=request.buffer.getbuffer(),
                ) as response:
                    await session.send(HTTP_RESPONSE_CREATE, serialize_response(packet["id"], response))
                    async for data, _ in response.content.iter_chunks():
                        await session.send(
                            HTTP_RESPONSE_CHUNK,
                            DataChunk({"id": packet["id"]}, data),
                        )
            await session.send(
                HTTP_RESPONSE_CLOSE,
                {"id": packet["id"]},
            )
        finally:
            del self.http_handles[packet["id"]]

    async def handle_http_request_send(self, session: Session, packet: DataChunk[RequestHandle]):
        if packet.meta["id"] not in self.http_handles:
            raise Exception(f"Request handle {packet.meta['id']} does not exist")
        handle = self.http_handles[packet.meta["id"]]
        if handle.session != session:
            raise PermissionDenied("Mismatched session on request handle")
        handle.buffer.write(packet.data)

    async def handle_http_request_close(self, session: Session, packet: RequestHandle):
        if packet["id"] not in self.http_handles:
            raise Exception(f"Request handle {packet['id']} does not exist")
        handle = self.http_handles[packet["id"]]
        if handle.session != session:
            raise PermissionDenied("Mismatched session on request handle")
        handle.close_future.set_result(handle.buffer)

    async def handle_ws_create(self, session: Session, packet: HttpRequest):
        if not session.permissions.has(HTTP_REQUEST_PERMISSION_ID):
            raise PermissionDenied(f"Missing HTTP request permission: {HTTP_REQUEST_PERMISSION_ID}")
        id = Identifier.from_key(packet["id"])
        if not session.is_app_id(id):
            raise Exception(f"Request id {packet['id']} is not a app id of {session.id.key()}")
        if packet["id"] in self.http_handles:
            raise Exception(f"Request id {packet['id']} already existing")

        try:
            url = URL(packet["url"])
            connect_future = Future[ClientWebSocketResponse]()
            request = WSHandle(session=session, socket=connect_future)
            await self.verify_port_allowed(session, url)
            self.ws_handles[packet["id"]] = request
            async with ClientSession(headers=USER_AGENT_HEADERS) as client:
                async with client.ws_connect(
                    url,
                    method=packet["method"],
                    headers=packet["header"],
                ) as socket:

                    async def _close(_):
                        if socket.closed:
                            return
                        await socket.close()

                    session.event.disconnected += _close

                    connect_future.set_result(socket)
                    await session.send(
                        WEBSOCKET_OPEN,
                        {"id": packet["id"], "protocol": socket.protocol, "url": str(socket._response.url)},
                    )
                    while not socket.closed:
                        received = await socket.receive()
                        if received.type in {WSMsgType.TEXT, WSMsgType.BINARY}:
                            data = received.data
                            if isinstance(data, str):
                                data = data.encode(encoding="utf-8")
                            if session.closed:
                                await socket.close()
                                break
                            await session.send(
                                WEBSOCKET_DATA,
                                DataChunk(
                                    WSDataMeta({"id": packet["id"], "type": received.type.value}),
                                    data,
                                ),
                            )
                        elif received.type == WSMsgType.CLOSE:
                            close: WSCloseCode = received.data
                            await session.send(
                                WEBSOCKET_CLOSE,
                                {"id": packet["id"], "code": close, "reason": received.extra},
                            )
                        else:
                            pass
        except aiohttp.ClientConnectorError:
            await session.send(
                WEBSOCKET_ERROR,
                {"id": packet["id"], "type": "ConnectionRefused", "reason": ""},
            )
        except NotAllowed as e:
            await session.send(
                WEBSOCKET_ERROR,
                {"id": packet["id"], "type": "ConnectionRefused", "reason": str(e)},
            )
        finally:
            self.ws_handles.pop(packet["id"], None)

    async def handle_ws_send(self, session: Session, packet: DataChunk[WSDataMeta]):
        if packet.meta["id"] not in self.ws_handles:
            raise Exception(f"Request handle {packet.meta['id']} does not exist")
        handle = self.ws_handles[packet.meta["id"]]
        if handle.session != session:
            raise PermissionDenied("Mismatched session on request handle")

        socket = await handle.socket
        if packet.meta["type"] == WSMsgType.TEXT:
            await socket.send_str(packet.data.decode("utf-8"))
        else:
            await socket.send_bytes(packet.data)

    async def handle_ws_close(self, session: Session, packet: WebSocketClose):
        if packet["id"] not in self.ws_handles:
            raise Exception(f"Request handle {packet['id']} does not exist")
        handle = self.ws_handles[packet["id"]]
        if handle.session != session:
            raise PermissionDenied("Mismatched session on request handle")
        socket = await handle.socket
        message = packet.get("reason") or ""
        await socket.close(
            code=packet.get("code", WSCloseCode.OK),
            message=message.encode(encoding="utf-8"),
        )



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from omu.api.permission import PermissionType
from omu.api.registry import RegistryPermissions, RegistryType
from omu.api.registry.extension import (
    REGISTRY_GET_ENDPOINT,
    REGISTRY_LISTEN_PACKET,
    REGISTRY_PERMISSION_ID,
    REGISTRY_REGISTER_PACKET,
    REGISTRY_UPDATE_PACKET,
    RegistryPacket,
)
from omu.api.registry.packets import RegisterPacket
from omu.errors import PermissionDenied
from omu.identifier import Identifier

from omuserver.session import Session

from .registry import Registry, ServerRegistry

if TYPE_CHECKING:
    from omuserver.server import Server

REGISTRY_PERMISSION = PermissionType(
    REGISTRY_PERMISSION_ID,
    {
        "level": "low",
        "name": {
            "ja": "データを保持",
            "en": "Application Data",
        },
        "note": {
            "ja": "アプリがデータを保持するために使われます",
            "en": "Used to store data for the app",
        },
    },
)


class RegistryExtension:
    def __init__(self, server: Server) -> None:
        self._server = server
        self.registries: dict[Identifier, ServerRegistry] = {}
        self._startup_registries: list[ServerRegistry] = []
        server.security.register_permission(REGISTRY_PERMISSION)
        server.packets.register(REGISTRY_REGISTER_PACKET, REGISTRY_LISTEN_PACKET, REGISTRY_UPDATE_PACKET)
        server.packets.bind(REGISTRY_REGISTER_PACKET, self.handle_register)
        server.packets.bind(REGISTRY_LISTEN_PACKET, self.handle_listen)
        server.packets.bind(REGISTRY_UPDATE_PACKET, self.handle_update)
        server.endpoints.bind(REGISTRY_GET_ENDPOINT, self.handle_get)
        server.event.start += self._on_start

    async def _on_start(self) -> None:
        for registry in self._startup_registries:
            await registry.load()
        self._startup_registries.clear()

    async def handle_register(self, session: Session, packet: RegisterPacket) -> None:
        registry = await self.get(packet.id)
        if not registry.id.is_subpath_of(session.app.id):
            msg = f"App {session.app.id=} not allowed to register {packet.id=}"
            raise PermissionDenied(msg)
        registry.permissions = packet.permissions

    async def handle_listen(self, session: Session, id: Identifier) -> None:
        await session.wait_ready()
        registry = await self.get_with_perm(id, session, lambda perms: [perms.all, perms.read])
        await registry.attach_session(session)

    async def handle_update(self, session: Session, packet: RegistryPacket) -> None:
        await session.wait_ready()
        registry = await self.get_with_perm(packet.id, session, lambda perms: [perms.all, perms.write])
        registry.store(packet.value)
        await registry.notify(session)

    async def handle_get(self, session: Session, id: Identifier) -> RegistryPacket:
        await session.wait_ready()
        registry = await self.get_with_perm(id, session, lambda perms: [perms.all, perms.read])
        return RegistryPacket(id, registry.value)

    async def get(self, id: Identifier) -> ServerRegistry:
        registry = self.registries.get(id)
        if registry is None:
            registry = ServerRegistry(
                server=self._server,
                id=id,
            )
            self.registries[id] = registry
            await registry.load()
        return registry

    async def get_with_perm(
        self, id: Identifier, session: Session, get_scope: Callable[[RegistryPermissions], list[Identifier | None]]
    ) -> ServerRegistry:
        registry = await self.get(id)
        self.verify_permission(registry, session, get_scope)
        return registry

    def verify_permission(
        self,
        registry: ServerRegistry,
        session: Session,
        get_scopes: Callable[[RegistryPermissions], list[Identifier | None]],
    ) -> None:
        if session.is_app_id(registry.id):
            return
        require_permissions = get_scopes(registry.permissions)
        if not session.permissions.has_any(filter(None, require_permissions)):
            msg = f"App {session.app.id=} not allowed to access {registry.id=}"
            raise PermissionDenied(msg)

    def register[T](
        self,
        registry_type: RegistryType[T],
    ) -> Registry[T]:
        registry = self.registries.get(registry_type.id)
        if registry is None:
            registry = ServerRegistry(
                server=self._server,
                id=registry_type.id,
                permissions=registry_type.permissions,
            )
            self.registries[registry_type.id] = registry
            self._startup_registries.append(registry)
        return Registry(
            registry,
            registry_type.default_value,
            registry_type.serializer,
        )



from __future__ import annotations

import time
from asyncio import Future
from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from omu.api.dashboard.extension import (
    DASHBOARD_ALLOWED_WEBVIEW_HOSTS,
    DASHBOARD_APP_INSTALL_ENDPOINT,
    DASHBOARD_APP_INSTALL_PACKET,
    DASHBOARD_APP_INSTALL_PERMISSION_ID,
    DASHBOARD_APP_UPDATE_PACKET,
    DASHBOARD_DRAG_DROP_READ_ENDPOINT,
    DASHBOARD_DRAG_DROP_STATE_PACKET,
    DASHBOARD_OPEN_APP_ENDPOINT,
    DASHBOARD_OPEN_APP_PACKET,
    DASHBOARD_OPEN_APP_PERMISSION_ID,
    DASHBOARD_PROMPT_CLEAR_BLOCKED,
    DASHBOARD_PROMPT_REQUEST,
    DASHBOARD_PROMPT_RESPONSE,
    DASHBOARD_SET_ENDPOINT,
    DASHBOARD_SPEECH_RECOGNITION,
    DASHBOARD_WEBVIEW_EVENT_PACKET,
    DashboardSetResponse,
    DragDropReadRequest,
    DragDropReadResponse,
    FileDragPacket,
    PortProcess,
    PromptRequest,
    PromptResponse,
    PromptResult,
    WebviewEventPacket,
)
from omu.api.dashboard.packets import (
    AppInstallResponse,
)
from omu.api.permission.permission import PermissionType
from omu.api.plugin.package_info import PackageInfo
from omu.api.server.extension import AppIndexRegistryMeta
from omu.app import App, AppType
from omu.errors import PermissionDenied
from omu.identifier import Identifier
from omu.result import Err, Ok
from psutil import AccessDenied
from yarl import URL

from omuserver.dependency import AppIndexRegistry
from omuserver.helper import find_processes_by_port
from omuserver.session import Session

from .permission import (
    DASHBOARD_APP_INSTALL_PERMISSION,
    DASHBOARD_DRAG_DROP_PERMISSION,
    DASHBOARD_OPEN_APP_PERMISSION,
    DASHBOARD_SET_PERMISSION,
    DASHBOARD_SPEECH_RECOGNITION_PERMISSION,
    DASHBOARD_WEBVIEW_PERMISSION,
)

if TYPE_CHECKING:
    from omuserver.server import Server


@dataclass(frozen=True, slots=True)
class PromptHandle:
    kind: str
    future: Future[PromptResult] = field(default_factory=Future)


class DashboardExtension:
    def __init__(self, server: Server) -> None:
        self.server = server
        server.packets.register(
            DASHBOARD_APP_INSTALL_PACKET,
            DASHBOARD_APP_UPDATE_PACKET,
            DASHBOARD_DRAG_DROP_STATE_PACKET,
            DASHBOARD_WEBVIEW_EVENT_PACKET,
            DASHBOARD_PROMPT_REQUEST,
            DASHBOARD_PROMPT_RESPONSE,
        )
        server.security.register_permission(
            DASHBOARD_SET_PERMISSION,
            DASHBOARD_OPEN_APP_PERMISSION,
            DASHBOARD_APP_INSTALL_PERMISSION,
            DASHBOARD_DRAG_DROP_PERMISSION,
            DASHBOARD_WEBVIEW_PERMISSION,
            DASHBOARD_SPEECH_RECOGNITION_PERMISSION,
        )
        server.packets.bind(DASHBOARD_DRAG_DROP_STATE_PACKET, self.handle_drag_drop_state)
        server.packets.bind(DASHBOARD_WEBVIEW_EVENT_PACKET, self.handle_webview_event)
        server.packets.bind(DASHBOARD_PROMPT_RESPONSE, self.handle_prompt_response)
        server.endpoints.bind(DASHBOARD_PROMPT_CLEAR_BLOCKED, self.clear_blocked_prompts)
        server.endpoints.bind(DASHBOARD_SET_ENDPOINT, self.handle_dashboard_set)
        server.endpoints.bind(DASHBOARD_OPEN_APP_ENDPOINT, self.handle_dashboard_open_app)
        server.endpoints.bind(DASHBOARD_APP_INSTALL_ENDPOINT, self.handle_dashboard_app_install)
        server.endpoints.bind(DASHBOARD_DRAG_DROP_READ_ENDPOINT, self.handle_drag_drop_read)
        self.allowed_webview_hosts = server.tables.register(DASHBOARD_ALLOWED_WEBVIEW_HOSTS)
        self.speech_recognition = server.registries.register(DASHBOARD_SPEECH_RECOGNITION)
        self.dashboard_session: Session | None = None
        self.dashboard_wait_future: Future[Session] | None = None
        self.prompt_requests: dict[str, PromptHandle] = {}
        self.prompt_keys: dict[str, PromptHandle] = {}
        self.blocked_prompts: set[str] = set()
        self.drag_drop_states: dict[str, Session] = {}
        self.drag_drop_read_requests: dict[str, Future[DragDropReadResponse]] = {}
        self.request_id = 0

    async def _handle_app_remove(self, apps: Mapping[str, App]):
        for app in apps.values():
            self.server.security.remove_app(app.id)

    async def wait_dashboard_ready(self) -> Session:
        if self.dashboard_session:
            return self.dashboard_session
        if self.dashboard_wait_future:
            return await self.dashboard_wait_future
        self.dashboard_wait_future = Future()
        return await self.dashboard_wait_future

    async def handle_dashboard_open_app(self, session: Session, app: App) -> None:
        if self.dashboard_session is None:
            raise ValueError("Dashboard session not set")
        if not session.permissions.has(DASHBOARD_OPEN_APP_PERMISSION_ID):
            raise PermissionDenied("Session does not have permission to open apps")
        await self.dashboard_session.send(DASHBOARD_OPEN_APP_PACKET, app)

    async def open_app(self, app: App) -> None:
        if self.dashboard_session is None:
            raise ValueError("Dashboard session not set")
        await self.dashboard_session.send(DASHBOARD_OPEN_APP_PACKET, app)

    async def handle_dashboard_set(self, session: Session, identifier: Identifier) -> DashboardSetResponse:
        if session.kind != AppType.DASHBOARD:
            raise PermissionDenied("Session is not a dashboard")
        self.dashboard_session = session
        session.event.disconnected += self._on_dashboard_disconnected
        if self.dashboard_wait_future:
            self.dashboard_wait_future.set_result(session)
            self.dashboard_wait_future = None
        return {"success": True}

    async def _on_dashboard_disconnected(self, session: Session) -> None:
        self.dashboard_session = None
        self.prompt_keys.clear()

    def ensure_dashboard_session(self, session: Session) -> bool:
        if session == self.dashboard_session:
            return True
        msg = f"Session {session} is not the dashboard session"
        raise PermissionDenied(msg)

    async def handle_prompt_response(self, session: Session, response: PromptResponse) -> None:
        self.ensure_dashboard_session(session)
        id = response["id"]

        if id not in self.prompt_requests:
            raise ValueError(f"Dashboard request with id {id} does not exist")

        handle = self.prompt_requests.pop(id)

        if handle.kind != response["kind"]:
            raise ValueError(
                f"The kind of Dashboard request with id {id} does not match {response['kind']} != {handle.kind}"
            )

        handle.future.set_result(response["result"])

    async def clear_blocked_prompts(self, session: Session, arg: None) -> None:
        self.blocked_prompts.clear()

    async def request_prompt(self, key: str, prompt: PromptRequest) -> bool:
        prompt_key = f"{prompt['kind']}-{key}"
        if prompt_key in self.blocked_prompts:
            return False
        if prompt_key in self.prompt_keys:
            return await self.prompt_keys[prompt_key].future == "accept"
        try:
            dashboard = await self.wait_dashboard_ready()
            if prompt["id"] in self.prompt_requests:
                raise ValueError(f"Permission request with id {prompt['id']} already exists")

            future = Future[PromptResult]()
            self.prompt_keys[prompt_key] = self.prompt_requests[prompt["id"]] = PromptHandle(prompt["kind"], future)

            await dashboard.send(
                DASHBOARD_PROMPT_REQUEST,
                prompt,
            )

            result = await future
            if result == "block":
                self.blocked_prompts.add(prompt_key)
            return result == "accept"
        finally:
            if prompt_key in self.prompt_keys:
                self.prompt_keys.pop(prompt_key)
            if prompt["id"] in self.prompt_requests:
                self.prompt_requests.pop(prompt["id"])

    async def request_permissions(self, app: App, permissions: list[PermissionType]) -> bool:
        return await self.request_prompt(
            app.id.key(),
            {
                "kind": "app/permissions",
                "id": self.get_next_request_id(),
                "app": app.to_json(),
                "permissions": list(map(PermissionType.to_json, permissions)),
            },
        )

    async def request_plugins(self, app: App, packages: list[PackageInfo]) -> bool:
        return await self.request_prompt(
            app.id.key(),
            {
                "kind": "app/plugins",
                "id": self.get_next_request_id(),
                "app": app.to_json(),
                "packages": packages,
            },
        )

    async def request_install(self, app: App, dependencies: dict[Identifier, App]) -> bool:
        return await self.request_prompt(
            app.id.key(),
            {
                "kind": "app/install",
                "id": self.get_next_request_id(),
                "app": app.to_json(),
                "dependencies": {k.key(): v.to_json() for k, v in dependencies.items()},
            },
        )

    async def handle_dashboard_app_install(self, session: Session, app: App) -> AppInstallResponse:
        if not session.permissions.has(DASHBOARD_APP_INSTALL_PERMISSION_ID):
            raise PermissionDenied("Session does not have permission to install apps")

        dependencies = await self.resolve_dependencies(app)

        if app.type == AppType.SERVICE and dependencies:
            raise Exception("Service apps cannot have dependencies")

        existing = self.server.security.apps.get(app.id)
        if existing:
            accepted = await self.notify_update_app(
                old_app=App.from_json(existing["app_json"]),
                new_app=app,
                dependencies=dependencies,
            )
        else:
            accepted = await self.request_install(app=app, dependencies=dependencies)

        if accepted:
            await self.server.server.apps.add(app, *dependencies.values())

        return AppInstallResponse(accepted=accepted)

    async def resolve_dependencies(self, app: App):
        indexes: dict[URL, AppIndexRegistry] = {}
        dependencies: dict[Identifier, App] = {}

        for dep_id_str, dep_specifier in (app.dependencies or {}).items():
            dep_id = Identifier.from_key(dep_id_str)

            match AppIndexRegistry.resolve_index_by_specifier(self.server, dep_id, dep_specifier):
                case Err(err):
                    raise Exception(f"Failed to resolve index URL for {dep_id}: {err}")
                case Ok(index_url):
                    pass

            if index_url not in indexes:
                match await AppIndexRegistry.try_fetch(index_url):
                    case Err(err):
                        raise Exception(f"Failed to fetch index from {index_url}: {err}")
                    case Ok(index):
                        indexes[index_url] = index

            index = indexes[index_url]
            if dep_id not in index.apps:
                raise Exception(f"Dependency {dep_id} not found in index {index_url}")
            dependencies[dep_id] = index.apps[dep_id]
        return dependencies

    async def notify_update_app(self, old_app: App, new_app: App, dependencies: dict[Identifier, App]) -> bool:
        return await self.request_prompt(
            old_app.id.key(),
            {
                "kind": "app/update",
                "id": self.get_next_request_id(),
                "old_app": old_app.to_json(),
                "new_app": new_app.to_json(),
                "dependencies": {k.key(): v.to_json() for k, v in dependencies.items()},
            },
        )

    async def notify_index_install(self, index_url: URL, meta: AppIndexRegistryMeta) -> bool:
        return await self.request_prompt(
            index_url.host or index_url.authority,
            {
                "kind": "index/install",
                "id": self.get_next_request_id(),
                "index_url": str(index_url),
                "meta": meta,
            },
        )

    async def notify_http_port_request(self, app: App, port: int) -> bool:
        processes: list[PortProcess] = []
        for process in find_processes_by_port(port):
            try:
                processes.append({"name": process.name(), "exe": process.exe(), "port": port})
            except AccessDenied:
                continue
        if not processes:
            return False
        return await self.request_prompt(
            f"{app.id.key()}-{port}",
            {
                "kind": "http/port",
                "id": self.get_next_request_id(),
                "app": app.to_json(),
                "processes": processes,
            },
        )

    async def handle_drag_drop_state(self, session: Session, packet: FileDragPacket):
        self.ensure_dashboard_session(session)
        id = Identifier.from_key(packet["app"]["id"])
        session = self.server.sessions.sessions[id]
        await session.send(DASHBOARD_DRAG_DROP_STATE_PACKET, packet)

    async def handle_drag_drop_read(self, session: Session, request: DragDropReadRequest) -> DragDropReadResponse:
        raise Exception("Not implemented")

    async def handle_webview_event(self, session: Session, packet: WebviewEventPacket):
        self.ensure_dashboard_session(session)
        target = self.server.sessions.find(packet.target)
        if target is None:
            return
        await target.send(DASHBOARD_WEBVIEW_EVENT_PACKET, packet)

    def get_next_request_id(self) -> str:
        self.request_id += 1
        return f"{self.request_id}-{time.time_ns()}"



from __future__ import annotations

import abc
import datetime
import hashlib
import json
import secrets
import urllib
import urllib.parse
from collections.abc import Iterable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal, TypedDict

from aiohttp import web
from omu.api.permission.permission import PermissionType
from omu.app import App, AppJson, AppType
from omu.errors import DisconnectReason, InvalidOrigin
from omu.helper import map_optional
from omu.identifier import Identifier
from omu.result import Err, Ok, Result

if TYPE_CHECKING:
    from omuserver.server import Server


class InputToken(str): ...


class HashedToken(str): ...


class PermissionHandle(abc.ABC):
    @abc.abstractmethod
    def grant(self, permission_id: Identifier) -> None: ...

    def grant_all(self, permission_ids: Iterable[Identifier]) -> None:
        for permission_id in permission_ids:
            self.grant(permission_id)

    @abc.abstractmethod
    def revoke(self, permission_id: Identifier) -> None: ...

    def revoke_all(self, permission_ids: Iterable[Identifier]) -> None:
        for permission_id in permission_ids:
            self.revoke(permission_id)

    @abc.abstractmethod
    def has(self, permission_id: Identifier) -> bool: ...

    def has_any(self, permission_ids: Iterable[Identifier]) -> bool:
        return any(map(self.has, permission_ids))

    def has_all(self, permission_ids: Iterable[Identifier]) -> bool:
        return all(map(self.has, permission_ids))


class FullPermissionHandle(PermissionHandle):
    def grant(self, permission_id: Identifier) -> None:
        pass

    def revoke(self, permission_id: Identifier) -> None:
        pass

    def has(self, permission_id: Identifier) -> bool:
        return True


@dataclass
class ParentPermissionHandle(PermissionHandle):
    manager: PermissionManager
    id: Identifier
    token: HashedToken

    def grant(self, permission_id: Identifier) -> None:
        permissions = set(self.manager.apps[self.id]["permissions"])
        if permission_id.key() in permissions:
            return
        permissions.add(permission_id.key())
        self.manager.apps[self.id]["permissions"] = list(permissions)
        self.manager.store()

    def revoke(self, permission_id: Identifier) -> None:
        permissions = self.manager.apps[self.id]["permissions"]
        if permission_id.key() not in permissions:
            return
        permissions.remove(permission_id.key())
        self.manager.store()

    def has(self, permission_id: Identifier) -> bool:
        permissions = self.manager.apps[self.id]["permissions"]
        return permission_id.key() in permissions


@dataclass
class ChildPermissionHandle(PermissionHandle):
    manager: PermissionManager
    id: Identifier
    token: HashedToken
    parent: PermissionHandle

    def grant(self, permission_id: Identifier) -> None:
        permissions = set(self.manager.apps[self.id]["permissions"])
        if permission_id.key() in permissions:
            return
        permissions.add(permission_id.key())
        self.manager.apps[self.id]["permissions"] = list(permissions)
        self.manager.store()

    def revoke(self, permission_id: Identifier) -> None:
        permissions = self.manager.apps[self.id]["permissions"]
        if permission_id.key() not in permissions:
            return
        permissions.remove(permission_id.key())
        self.manager.store()

    def has(self, permission_id: Identifier) -> bool:
        permissions = self.manager.apps[self.id]["permissions"]
        if permission_id.key() not in permissions:
            return False
        return self.parent.has(permission_id)


class AppEntry(TypedDict):
    token: HashedToken
    salt: str
    identifier: str
    app_json: AppJson
    type: Literal["app", "remote", "plugin", "dashboard"]
    permissions: list[str]

    # Statistics
    created_at: float
    last_used_at: float
    used_count: int


@dataclass
class PermissionManager:
    server: Server
    apps: dict[Identifier, AppEntry]
    permissions: dict[Identifier, PermissionType]
    frame_tokens: dict[str, str] = field(default_factory=dict)
    plugin_tokens: list[str] = field(default_factory=list)

    @classmethod
    def load(cls, server: Server) -> PermissionManager:
        security_dir = server.directories.get("security")
        apps_path = security_dir / "apps.json"
        permissions_path = security_dir / "permissions.json"
        if apps_path.exists():
            apps = dict(json.loads(apps_path.read_text(encoding="utf-8")))
        else:
            apps = {}
        if permissions_path.exists():
            permissions = dict(json.loads(permissions_path.read_text(encoding="utf-8")))
        else:
            permissions = {}
        return PermissionManager(
            server,
            {Identifier.from_key(k): v for k, v in apps.items()},
            {Identifier.from_key(k): PermissionType.from_json(v) for k, v in permissions.items()},
        )

    def store(self):
        security_dir = self.server.directories.get("security")
        apps_json = json.dumps({k.key(): v for k, v in self.apps.items()}, ensure_ascii=False)
        permissions_json = json.dumps({k.key(): v.to_json() for k, v in self.permissions.items()}, ensure_ascii=False)
        (security_dir / "apps.json").write_text(apps_json, encoding="utf-8")
        (security_dir / "permissions.json").write_text(permissions_json, encoding="utf-8")

    def register_permission(self, *permissions: PermissionType, overwrite: bool = True) -> None:
        existing_permissions = list(filter(lambda perm: perm.id in self.permissions, permissions))
        if existing_permissions and not overwrite:
            raise ValueError(f"Permissions already registered: {', '.join(map(str, existing_permissions))}")
        for permission in permissions:
            self.permissions[permission.id] = permission

    def _random_token(self, length=32):
        return secrets.token_urlsafe(length)[:length]

    def generate_frame_token(self, url: str) -> str:
        token = self._random_token()
        self.frame_tokens[token] = url
        return token

    def generate_plugin_token(self) -> str:
        token = self._random_token()
        self.plugin_tokens.append(token)
        return token

    def generate_app_token(self, app: App) -> tuple[PermissionHandle, InputToken]:
        salt = self._random_token()
        input_token = InputToken(self._random_token())
        hash_obj = hashlib.sha256()
        hash_obj.update((salt + input_token).encode("utf-8"))
        hashed_token = HashedToken(hash_obj.hexdigest())
        entry: AppEntry = {
            "token": hashed_token,
            "salt": salt,
            "identifier": app.id.key(),
            "app_json": app.to_json(),
            "type": "app",
            "permissions": [],
            "created_at": datetime.datetime.now().timestamp(),
            "last_used_at": 0,
            "used_count": 0,
        }
        existing = self.apps.get(app.id)
        if existing is not None:
            entry["permissions"] = existing["permissions"]
            if app.metadata is None:
                entry["app_json"]["metadata"] = existing["app_json"].get("metadata")
        self.apps[app.id] = entry
        self.store()
        return (ParentPermissionHandle(self, app.id, hashed_token), input_token)

    def remove_app(self, id: Identifier):
        if id not in self.apps:
            return
        del self.apps[id]
        self.store()

    def match_token(self, entry: AppEntry, token: InputToken) -> Result[HashedToken, str]:
        expected_token = entry["token"]
        hash_obj = hashlib.sha256()
        hash_obj.update((entry["salt"] + token).encode("utf-8"))
        hashed_token = hash_obj.hexdigest()
        token_matched = secrets.compare_digest(hashed_token, expected_token)
        if not token_matched:
            return Err("Token does not match")
        return Ok(expected_token)

    def verify_frame_token(self, request: web.Request) -> Result[None, DisconnectReason]:
        query = request.query
        frame_token = query.get("frame_token")
        url = query.get("url")
        origin = request.headers.get("Origin")
        if frame_token is None:
            return Err(InvalidOrigin("Missing frame token"))
        if origin is None:
            return Err(InvalidOrigin("Missing origin"))
        if url is None:
            return Err(InvalidOrigin("Missing url"))

        parsed_url = urllib.parse.unquote(url)
        token_matched = self.frame_tokens[frame_token] == parsed_url

        if not token_matched:
            return Err(InvalidOrigin("Invalid frame token"))
        return Ok(None)

    async def verify_app(self, app: App, token: InputToken) -> Result[tuple[PermissionHandle, InputToken], str]:
        if app.type == AppType.DASHBOARD:
            token_matched = self.server.config.dashboard_token == token
            if not token_matched:
                return Err("Invalid token")
            return Ok((FullPermissionHandle(), token))
        if app.type == AppType.PLUGIN:
            token_matched = token in self.plugin_tokens
            if not token_matched:
                return Err("Invalid token")
            return Ok((FullPermissionHandle(), token))
        return await self.verify_generic_app(app, token)

    async def verify_generic_app(self, app: App, token: InputToken) -> Result[tuple[PermissionHandle, InputToken], str]:
        found_entry = self.apps.get(app.id)
        if found_entry is None:
            return Err(f"App {app.id} not found")
        token_result = self.match_token(found_entry, token)
        if token_result.is_err is True:
            return Err(f"Token mismatch for app {app.id}: {token_result.err}")

        match_result = self.match_app(found_entry["app_json"], app)
        if match_result.is_err is True:
            if found_entry["app_json"].get("metadata") is not None:
                dependencies = await self.server.dashboard.resolve_dependencies(app)
                accepted = await self.server.dashboard.notify_update_app(
                    old_app=App.from_json(found_entry["app_json"]),
                    new_app=app,
                    dependencies=dependencies,
                )
                if not accepted:
                    return Err(f"App data mismatch for app {app.id} and not accepted this change: {match_result.err}")
                await self.server.server.apps.add(*dependencies.values())
            self.apps[app.id]["app_json"] = app.to_json()
            self.store()

        if app.parent_id:
            parent_entry = self.apps.get(app.parent_id)
            if parent_entry is None:
                return Err(f"Parent app {app.parent_id} not found")
            if parent_entry["app_json"].get("parent_id"):
                return Err("Child apps cannot have children")
            parent = ParentPermissionHandle(self, app.id, parent_entry["token"])
            child = ChildPermissionHandle(self, app.id, token_result.value, parent)
            return Ok((child, token))
        else:
            child = ParentPermissionHandle(self, app.id, token_result.value)
            return Ok((child, token))

    def match_app(self, existing: AppJson, new_app: App) -> Result[..., str]:
        if existing["id"] != new_app.id.key():
            return Err("App ID does not match")
        if existing.get("parent_id") != map_optional(new_app.parent_id, Identifier.key):
            return Err("App parent ID does not match")
        if existing.get("url") != new_app.url:
            return Err("App URL does not match")
        if existing.get("type") != new_app.type.value:
            return Err("App type does not match")
        if existing.get("dependencies") != new_app.dependencies:
            return Err("App type does not match")
        if json.dumps(existing.get("metadata")) != json.dumps(new_app.metadata):
            return Err("App metadata does not match")
        needed_dependencies = list(
            filter(lambda id: Identifier.from_key(id) not in self.apps, (new_app.dependencies or {}).keys())
        )
        if needed_dependencies:
            return Err(f"App's dependencies are not satisfied: {', '.join(needed_dependencies)}")
        return Ok(...)



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import TYPE_CHECKING

from loguru import logger
from omu.api.endpoint.extension import (
    ENDPOINT_INVOKE_PACKET,
    ENDPOINT_INVOKED_PACKET,
    ENDPOINT_REGISTER_PACKET,
    ENDPOINT_RESPONSE_PACKET,
    EndpointType,
)
from omu.api.endpoint.packets import (
    EndpointInvokedPacket,
    EndpointInvokePacket,
    EndpointRegisterPacket,
    EndpointResponsePacket,
    InvokedParams,
    ResponseParams,
)
from omu.api.extension import EXT_NAMESPACE
from omu.app import AppType
from omu.errors import PermissionDenied
from omu.helper import Coro
from omu.identifier import Identifier
from omu.network.packet.packet_types import DisconnectType

from omuserver.session import Session

if TYPE_CHECKING:
    from omuserver.server import Server


class Endpoint(abc.ABC):
    @property
    @abc.abstractmethod
    def id(self) -> Identifier: ...

    @property
    @abc.abstractmethod
    def permission(self) -> Identifier | None: ...

    @abc.abstractmethod
    async def call(self, data: EndpointInvokedPacket, session: Session) -> Session | None: ...


class SessionEndpoint(Endpoint):
    def __init__(
        self,
        session: Session,
        id: Identifier,
        permission: Identifier | None,
    ) -> None:
        self._session = session
        self._id = id
        self._permission = permission

    @property
    def id(self) -> Identifier:
        return self._id

    @property
    def permission(self) -> Identifier | None:
        return self._permission

    async def call(self, data: EndpointInvokedPacket, session: Session) -> Session:
        if self._session.closed:
            raise RuntimeError(f"Session {self._session.app.key()} already closed")
        await self._session.send(ENDPOINT_INVOKED_PACKET, data)
        return self._session


class ServerEndpoint[Req, Res](Endpoint):
    def __init__(
        self,
        server: Server,
        endpoint: EndpointType[Req, Res],
        callback: Coro[[Session, Req], Res],
        permission: Identifier | None = None,
    ) -> None:
        self._server = server
        self._endpoint = endpoint
        self._callback = callback
        self._permission = permission

    @property
    def id(self) -> Identifier:
        return self._endpoint.id

    @property
    def permission(self) -> Identifier | None:
        return self._permission

    async def call(self, data: EndpointInvokedPacket, session: Session) -> None:
        if session.closed:
            raise RuntimeError("Session already closed")
        try:
            req = self._endpoint.request_serializer.deserialize(data.buffer)
            res = await self._callback(session, req)
            serialized = self._endpoint.response_serializer.serialize(res)
            await session.send(
                ENDPOINT_RESPONSE_PACKET,
                EndpointResponsePacket(
                    ResponseParams(
                        id=data.params.id,
                        key=data.params.key,
                        error=None,
                    ),
                    buffer=serialized,
                ),
            )
        except Exception as e:
            await session.send(
                ENDPOINT_RESPONSE_PACKET,
                EndpointResponsePacket(
                    ResponseParams(
                        id=data.params.id,
                        key=data.params.key,
                        error=repr(e),
                    ),
                    buffer=b"",
                ),
            )
            raise e
        return None


@dataclass(frozen=True, slots=True)
class EndpointInvoke:
    caller: Session
    session: Session
    invoke: EndpointInvokedPacket


class EndpointExtension:
    def __init__(self, server: Server) -> None:
        self._server = server
        self._endpoints: dict[Identifier, Endpoint] = {}
        self._invokes: dict[tuple[Identifier, int], EndpointInvoke] = {}
        server.packets.register(
            ENDPOINT_REGISTER_PACKET,
            ENDPOINT_INVOKE_PACKET,
            ENDPOINT_INVOKED_PACKET,
            ENDPOINT_RESPONSE_PACKET,
        )
        server.packets.bind(ENDPOINT_REGISTER_PACKET, self.handle_register)
        server.packets.bind(ENDPOINT_INVOKE_PACKET, self.handle_invoke)
        server.packets.bind(ENDPOINT_RESPONSE_PACKET, self.handle_response)

    async def handle_register(self, session: Session, packet: EndpointRegisterPacket) -> None:
        for id, permission in packet.endpoints.items():
            if not self.is_endpoint_binding_allowed(session, id):
                msg = f"App {session.app.key()} not allowed to register endpoint {id}"
                raise PermissionDenied(msg)
            self._endpoints[id] = SessionEndpoint(
                session=session,
                id=id,
                permission=permission,
            )

    def is_endpoint_binding_allowed(self, session: Session, id: Identifier) -> bool:
        if id.is_subpath_of(session.app.id):
            return True
        if session.kind == AppType.DASHBOARD and id.namespace == EXT_NAMESPACE:
            return True
        return False

    def bind[Req, Res](
        self,
        type: EndpointType[Req, Res],
        callback: Coro[[Session, Req], Res],
    ) -> None:
        if type.id in self._endpoints:
            raise ValueError(f"Endpoint {type.id.key()} already bound")
        endpoint = ServerEndpoint(
            server=self._server,
            endpoint=type,
            callback=callback,
            permission=type.permission_id,
        )
        self._endpoints[type.id] = endpoint

    def verify_permission(self, endpoint: Endpoint, session: Session):
        if session.is_app_id(endpoint.id):
            return
        if endpoint.permission and session.permissions.has(endpoint.permission):
            return
        error = f"{session.app.key()} tried to call endpoint {endpoint.id} without permission {endpoint.permission}"
        logger.warning(error)
        raise PermissionDenied(error)

    async def handle_invoke(self, session: Session, packet: EndpointInvokePacket) -> None:
        endpoint = self._endpoints.get(packet.params.id)
        if endpoint is None:
            logger.warning(f"{session.app.key()} tried to call unknown endpoint {packet.params.id}")
            await session.send(
                ENDPOINT_RESPONSE_PACKET,
                EndpointResponsePacket(
                    ResponseParams(
                        id=packet.params.id,
                        key=packet.params.key,
                        error=f"Endpoint {packet.params.id} not found",
                    ),
                    buffer=b"",
                ),
            )
            return
        self.verify_permission(endpoint, session)
        invoke = EndpointInvokedPacket(
            InvokedParams(
                id=packet.params.id,
                key=packet.params.key,
                caller=session.app.id,
            ),
            buffer=packet.buffer,
        )
        endpoint_session = await endpoint.call(invoke, session)
        if endpoint_session is None:
            return
        key = (packet.params.id, packet.params.key)
        self._invokes[key] = EndpointInvoke(
            caller=session,
            session=endpoint_session,
            invoke=invoke,
        )

    async def handle_response(self, session: Session, packet: EndpointResponsePacket) -> None:
        key = (packet.params.id, packet.params.key)
        invoke = self._invokes.get(key)
        if invoke is None:
            logger.warning(
                f"{session.app.key()} tried to handle response for unknown call {key} ({packet.params.id.key()})"
            )
            return
        if session != invoke.session:
            await session.disconnect(
                DisconnectType.PERMISSION_DENIED,
                f"Session mismatch for call {key} ({packet.params.id.key()}): {session.app} != {invoke.session.app}",
            )
            return
        del self._invokes[key]
        await invoke.caller.send(ENDPOINT_RESPONSE_PACKET, packet)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

import aiohttp
from omu.api.server.extension import AppIndexRegistryMeta
from omu.app import App, AppJson, DependencySpecifier
from omu.identifier import Identifier
from omu.result import Err, Ok, Result
from yarl import URL

from omuserver.consts import USER_AGENT_HEADERS

if TYPE_CHECKING:
    from omuserver.server import Server


class InstallRequest(TypedDict):
    index: str
    id: str


class AppIndexRegistryJSON(TypedDict):
    id: str
    apps: dict[str, AppJson]
    meta: AppIndexRegistryMeta


@dataclass(frozen=True, slots=True)
class AppIndexRegistry:
    id: Identifier
    apps: dict[Identifier, App]
    meta: AppIndexRegistryMeta

    @staticmethod
    def from_json(json: AppIndexRegistryJSON) -> AppIndexRegistry:
        id = Identifier.from_key(json["id"])
        apps: dict[Identifier, App] = {}
        for id_str, app_json in json["apps"].items():
            app_id = Identifier.from_key(id_str)
            app = App.from_json(app_json)
            if not app_id.is_namepath_equal(app.id, 0):
                raise AssertionError(f"App ID does not match the ID in the index. {app.id} != {app_id}")
            apps[app_id] = app
        return AppIndexRegistry(id=id, apps=apps, meta=json["meta"])

    @staticmethod
    def resolve_index_by_specifier(
        server: Server, id: Identifier, specifier: DependencySpecifier | str
    ) -> Result[URL, str]:
        if not isinstance(specifier, str):
            index_url = specifier.get("index")
            if index_url is not None:
                mapped_url = server.network.get_mapped_url(URL(index_url))
                index_id = Identifier.from_url(mapped_url)
                if not index_id.is_namepath_equal(id, 0):
                    return Err(f"Invalid Index ID: {index_id.namespace} != {id.namespace}")
                return Ok(URL(index_url))
        url = id.into_url()
        url.path = "/_omuapps.json"
        return Ok(url)

    @classmethod
    async def try_fetch(cls, url: URL) -> Result[AppIndexRegistry, str]:
        async with aiohttp.ClientSession(headers=USER_AGENT_HEADERS) as client:
            try:
                async with client.get(
                    url,
                    headers={"Content-Type": "application/json"},
                ) as response:
                    response.raise_for_status()
                    data: AppIndexRegistryJSON = await response.json()
                    try:
                        index = AppIndexRegistry.from_json(data)
                        return Ok(index)
                    except AssertionError as e:
                        return Err(str(e))

            except aiohttp.ClientError as e:
                return Err(str(e))



from __future__ import annotations

from typing import TYPE_CHECKING

from loguru import logger
from omu.api.logger.extension import LOGGER_LISTEN_PACKET, LOGGER_LOG_PACKET, LOGGER_LOG_PERMISSION_ID
from omu.api.logger.packets import LogMessage, LogPacket
from omu.errors import PermissionDenied
from omu.identifier import Identifier

from omuserver.session import Session

from .permissions import LOGGER_LOG_PERMISSION

if TYPE_CHECKING:
    from omuserver.server import Server


class LoggerExtension:
    def __init__(self, server: Server):
        server.security.register_permission(LOGGER_LOG_PERMISSION)
        server.packets.register(LOGGER_LOG_PACKET, LOGGER_LISTEN_PACKET)
        server.packets.bind(LOGGER_LOG_PACKET, self.handle_log)
        server.packets.bind(LOGGER_LISTEN_PACKET, self.handle_listen)
        self.listeners: dict[Identifier, set[Session]] = {}

    async def broadcast(self, id: Identifier, message: LogMessage) -> None:
        packet = LogPacket(id=id, message=message)
        for session in self.listeners.get(id, []):
            await session.send(LOGGER_LOG_PACKET, packet)

    async def handle_log(self, session: Session, packet: LogPacket) -> None:
        if not session.permissions.has(LOGGER_LOG_PERMISSION_ID):
            raise PermissionDenied("You do not have permission to log")
        logger.info(f"{packet.id}: {packet.message}")

    async def handle_listen(self, session: Session, id: Identifier) -> None:
        if not session.permissions.has(LOGGER_LOG_PERMISSION_ID):
            raise PermissionDenied("You do not have permission to listen to logs")
        if id not in self.listeners:
            self.listeners[id] = set()
        self.listeners[id].add(session)



from omu.api.dashboard import (
    DASHBOARD_APP_INSTALL_PERMISSION_ID,
    DASHBOARD_DRAG_DROP_PERMISSION_ID,
    DASHBOARD_OPEN_APP_PERMISSION_ID,
    DASHBOARD_SET_PERMISSION_ID,
    DASHBOARD_SPEECH_RECOGNITION_PERMISSION_ID,
    DASHBOARD_WEBVIEW_PERMISSION_ID,
)
from omu.api.permission import PermissionType

DASHBOARD_SET_PERMISSION = PermissionType(
    DASHBOARD_SET_PERMISSION_ID,
    {
        "level": "low",
        "name": {
            "ja": "管理者権限",
            "en": "Manage the dashboard",
        },
        "note": {
            "ja": "アプリが権限の管理やユーザーに確認を行うために使われます",
            "en": "Used by apps to manage permissions and confirm users",
        },
    },
)
DASHBOARD_OPEN_APP_PERMISSION = PermissionType(
    DASHBOARD_OPEN_APP_PERMISSION_ID,
    {
        "level": "medium",
        "name": {
            "ja": "アプリを開く",
            "en": "Open an app",
        },
        "note": {
            "ja": "インストールされているアプリを起動するために使われます",
            "en": "Used to start an installed app",
        },
    },
)
DASHBOARD_APP_INSTALL_PERMISSION = PermissionType(
    DASHBOARD_APP_INSTALL_PERMISSION_ID,
    {
        "level": "high",
        "name": {
            "ja": "アプリを追加",
            "en": "Install an app",
        },
        "note": {
            "ja": "新しくアプリを追加するために使われます",
            "en": "Used to install an app",
        },
    },
)
DASHBOARD_DRAG_DROP_PERMISSION = PermissionType(
    DASHBOARD_DRAG_DROP_PERMISSION_ID,
    {
        "level": "low",
        "name": {
            "ja": "ファイルのドラッグドロップ",
            "en": "Get File Drag Drop Information",
        },
    },
)
DASHBOARD_WEBVIEW_PERMISSION = PermissionType(
    DASHBOARD_WEBVIEW_PERMISSION_ID,
    {
        "level": "high",
        "name": {
            "ja": "外部サイトの認証情報を要求",
            "en": "Use external sites login information",
        },
        "note": {
            "ja": "許可したサイトへのすべての操作が可能になります",
            "en": "All operations on the site will be enabled.",
        },
    },
)
DASHBOARD_SPEECH_RECOGNITION_PERMISSION = PermissionType(
    DASHBOARD_SPEECH_RECOGNITION_PERMISSION_ID,
    {
        "level": "medium",
        "name": {
            "ja": "音声認識を使用",
            "en": "Use Speech Recognition",
        },
    },
)



from __future__ import annotations

from asyncio import Future
from collections import defaultdict
from typing import TYPE_CHECKING, Final

from loguru import logger
from omu.address import get_lan_ip
from omu.api.session.extension import (
    GENERATE_TOKEN_ENDPOINT_TYPE,
    REMOTE_APP_REQUEST_ENDPOINT_TYPE,
    SESSION_CONNECTED_PACKET_TYPE,
    SESSION_DISCONNECTED_PACKET_TYPE,
    SESSION_OBSERVE_PACKET_TYPE,
    SESSION_REQUIRE_PACKET_TYPE,
    SESSION_TABLE_TYPE,
    SESSIONS_READ_PERMISSION_ID,
    GenerateTokenPayload,
    GenerateTokenResponse,
    RemoteAppRequestPayload,
    RequestRemoteAppResponse,
)
from omu.app import App, AppType
from omu.errors import PermissionDenied
from omu.identifier import Identifier
from omu.network.packet.packet_types import DisconnectType

if TYPE_CHECKING:
    from omuserver.server import Server
from omuserver.session import Session, TaskPriority

from .permissions import GENERATE_TOKEN_PERMISSION, REMOTE_APP_REQUEST_PERMISSION, SESSIONS_READ_PERMISSION


class SessionWaiter:
    def __init__(self, ids: list[Identifier]):
        self.future = Future()
        self.ids = ids


class SessionExtension:
    def __init__(self, server: Server):
        self.server = server
        server.security.register_permission(
            REMOTE_APP_REQUEST_PERMISSION,
            SESSIONS_READ_PERMISSION,
            GENERATE_TOKEN_PERMISSION,
        )
        self.sessions: Final[dict[Identifier, Session]] = {}
        self.session_table = server.tables.register(SESSION_TABLE_TYPE)
        self.session_waiters: dict[Identifier, list[SessionWaiter]] = defaultdict(list)
        self.session_observers: dict[Identifier, list[Session]] = defaultdict(list)
        server.endpoints.bind(REMOTE_APP_REQUEST_ENDPOINT_TYPE, self.handle_remote_app_request)
        server.endpoints.bind(GENERATE_TOKEN_ENDPOINT_TYPE, self.handle_generate_token)
        server.packets.register(
            SESSION_REQUIRE_PACKET_TYPE,
            SESSION_OBSERVE_PACKET_TYPE,
            SESSION_CONNECTED_PACKET_TYPE,
            SESSION_DISCONNECTED_PACKET_TYPE,
        )
        server.packets.bind(SESSION_REQUIRE_PACKET_TYPE, self.handle_require)
        server.packets.bind(SESSION_OBSERVE_PACKET_TYPE, self.handle_observe)
        server.event.start += self.on_start

    async def on_start(self):
        await self.session_table.clear()

    async def handle_require(self, session: Session, app_ids: list[Identifier]) -> None:
        if len(app_ids) == 0:
            return

        async def task():
            for app_id in app_ids:
                existing_session = self.server.sessions.find(app_id)
                if existing_session:
                    await existing_session.wait_ready()
                    continue
                waiter = SessionWaiter(app_ids)
                for app_id in app_ids:
                    self.session_waiters[app_id].append(waiter)
                await waiter.future

        session.add_task(
            task,
            name="session_require",
            detail=f"Requiring sessions: {', '.join(map(Identifier.key, app_ids))}",
            priority=TaskPriority.AFTER_PLUGIN,
        )

    async def handle_observe(self, session: Session, app_ids: list[Identifier]) -> None:
        if not session.permissions.has(SESSIONS_READ_PERMISSION_ID):
            raise PermissionDenied(f"Pemission {SESSIONS_READ_PERMISSION_ID} required to observe session")
        for app_id in app_ids:
            self.session_observers[app_id].append(session)

            def on_disconnect(session, app_id=app_id):
                if session in self.session_observers[app_id]:
                    self.session_observers[app_id].remove(session)

            session.event.disconnected.listen(on_disconnect)

    async def handle_generate_token(self, session: Session, request: GenerateTokenPayload) -> GenerateTokenResponse:
        requested_app = App.from_json(request["app"])
        if session.app.parent_id:
            return {"type": "error", "message": "Only parent apps can generate tokens"}
        if session.kind != AppType.DASHBOARD:
            if not requested_app.id.is_subpath_of(session.id):
                return {"type": "error", "message": "You can only generate tokens for sibling apps"}
            if requested_app.parent_id is None:
                return {"type": "error", "message": "You can only generate tokens for child apps"}
            if requested_app.parent_id != session.app.id:
                return {"type": "error", "message": "You can only generate tokens for your own child apps"}
        permissions = list(map(Identifier.from_key, request.get("permissions", [])))
        if not session.permissions.has_all(permissions):
            missing_permissions = [p for p in permissions if not session.permissions.has(p)]
            return {
                "type": "error",
                "message": "You don't have permission to grant some of the requested permissions "
                + ", ".join(map(str, missing_permissions)),
            }
        handle, token = self.server.security.generate_app_token(requested_app)
        handle.grant_all(permissions)
        return {
            "type": "success",
            "token": token,
        }

    async def handle_remote_app_request(
        self, session: Session, request: RemoteAppRequestPayload
    ) -> RequestRemoteAppResponse:
        requested_app = App.from_json(request["app"])
        if session.app.parent_id:
            return {"type": "error", "message": "Only parent apps can generate tokens"}
        if session.kind != AppType.DASHBOARD:
            if not requested_app.id.is_subpath_of(session.id):
                return {"type": "error", "message": "You can only generate tokens for sibling apps"}
            if requested_app.parent_id is None:
                return {"type": "error", "message": "You can only generate tokens for child apps"}
            if requested_app.parent_id != session.app.id:
                return {"type": "error", "message": "You can only generate tokens for your own child apps"}
            if requested_app.url is None:
                return {"type": "error", "message": "Requested app must have a URL"}
        permissions = list(map(Identifier.from_key, request["permissions"]))
        if not session.permissions.has_all(permissions):
            missing_permissions = [p for p in permissions if not session.permissions.has(p)]
            return {
                "type": "error",
                "message": "You don't have permission to grant some of the requested permissions "
                + ", ".join(map(str, missing_permissions)),
            }
        handle, token = self.server.security.generate_app_token(requested_app)
        handle.grant_all(permissions)
        return {
            "type": "success",
            "token": token,
            "lan_ip": get_lan_ip(),
        }

    async def process_new(self, session: Session) -> None:
        logger.info(f"Connected: {session.app.key()}")
        existing_session = self.find(session.app.id)
        if existing_session:
            logger.warning(f"Session {session.app} already connected")
            await existing_session.disconnect(
                DisconnectType.ANOTHER_CONNECTION,
                f"Another connection from {session.app}",
            )
        self.sessions[session.app.id] = session
        session.event.disconnected += self.handle_disconnection
        await self.server.server.apps.update(session.app)
        await self.session_table.add(session.app)
        unlisten = session.event.ready.listen(self.on_session_ready)
        session.event.disconnected.listen(lambda _: unlisten())
        await session.listen()

    def iter(self):
        return self.sessions.values()

    async def on_disconnected(self, session: Session) -> None:
        logger.info(f"Disconnected: {session.app.key()}")
        await self.session_table.remove(session.app)

        for observer in self.session_observers.get(session.app.id, []):
            if session.closed:
                continue
            await observer.send(SESSION_DISCONNECTED_PACKET_TYPE, session.app)

    async def handle_disconnection(self, session: Session) -> None:
        if session.app.id not in self.sessions:
            return
        del self.sessions[session.app.id]

    def find(self, id: Identifier) -> Session | None:
        return self.sessions.get(id)

    async def on_session_ready(self, session: Session) -> None:
        for waiter in self.session_waiters.get(session.app.id, []):
            if session.app.id in waiter.ids:
                waiter.ids.remove(session.app.id)
            if len(waiter.ids) == 0:
                waiter.future.set_result(True)
                self.session_waiters.pop(session.app.id)
        for observer in self.session_observers.get(session.app.id, []):
            await observer.send(
                SESSION_CONNECTED_PACKET_TYPE,
                session.app,
            )



from __future__ import annotations

import asyncio
import importlib.metadata
import sys
import tempfile
from collections.abc import Mapping
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Protocol,
)

import aiohttp
import uv
from loguru import logger
from omu.api.plugin import PackageInfo, PluginPackageInfo
from omu.plugin import InstallContext, Plugin
from omu.result import Err, Ok, Result, is_err
from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from omuserver.consts import USER_AGENT_HEADERS
from omuserver.helper import normalize_package_name

from .entry import PluginEntry
from .instance import PluginInstance

if TYPE_CHECKING:
    from omuserver.server import Server


class PluginModule(Protocol):
    plugin: Plugin


class RequiredVersionTooOld(Exception):
    pass


class DependencyResolver:
    def __init__(self, server: Server) -> None:
        self.server = server
        self._dependencies: dict[str, SpecifierSet] = {}
        self._packages_distributions: Mapping[str, importlib.metadata.Distribution] = {}
        self._package_info_cache: dict[str, PackageInfo] = {}
        self._distributions_change_marked = True
        self.find_packages_distributions()

    async def fetch_package_info(self, package: str) -> PackageInfo:
        if package in self._package_info_cache:
            return self._package_info_cache[package]
        async with aiohttp.ClientSession(headers=USER_AGENT_HEADERS) as session:
            async with session.get(f"https://pypi.org/pypi/{package}/json") as response:
                package_info = PackageInfo(await response.json())
                self._package_info_cache[package] = package_info
                return package_info

    async def get_installed_package_info(self, package: str) -> PluginPackageInfo | None:
        try:
            package_info = importlib.metadata.distribution(package)
        except importlib.metadata.PackageNotFoundError:
            return None
        return PluginPackageInfo(
            package=package_info.name,
            version=package_info.version,
        )

    def format_dependencies(self, dependencies: Mapping[str, SpecifierSet | None]) -> list[str]:
        args = []
        for dependency, specifier in dependencies.items():
            if specifier is not None:
                args.append(f"{dependency}{specifier}")
            else:
                args.append(dependency)
        return args

    async def install_requirements(self, requirements: dict[str, SpecifierSet]) -> Result[None, str]:
        if len(requirements) == 0:
            return Ok(None)
        with tempfile.NamedTemporaryFile(mode="wb", delete=True) as req_file:
            dependency_lines = self.format_dependencies(requirements)
            req_file.write("\n".join(dependency_lines).encode("utf-8"))
            req_file.flush()
            index_args: list[str] = []
            if self.server.config.index_url:
                index_args = ["--extra-index-url", self.server.config.index_url]
            process = await asyncio.create_subprocess_exec(
                uv.find_uv_bin(),
                "pip",
                "install",
                "--upgrade",
                "-r",
                req_file.name,
                "--python",
                sys.executable,
                *index_args,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await process.communicate()
        if process.returncode != 0:
            return Err(f"Failed to install dependencies: stdout={stdout.decode()} stderr={stderr.decode()}")
        logger.info(f"Ran uv command: {(stdout or stderr).decode()}")
        return Ok(None)

    def is_package_satisfied(self, package: str, specifier: SpecifierSet | None) -> Result[..., str]:
        package_info = self._packages_distributions.get(package)
        if package_info is None:
            return Err(f"Package {package} not installed")
        if specifier is None:
            return Ok(...)
        installed_version = Version(package_info.version)
        if installed_version not in specifier:
            return Err(
                f"Installed version {installed_version} does not satify required package {package} with specifier {specifier}"
            )
        return Ok(...)

    def is_requirements_satisfied(self, requirements: dict[str, SpecifierSet | None]) -> Result[..., str]:
        for package, specifier in requirements.items():
            match self.is_package_satisfied(package, specifier):
                case Err(err):
                    return Err(f"Requirement not satisfied: {err}")
        return Ok(...)

    def _get_minimum_version(self, specifier: SpecifierSet) -> Version:
        minimum_version = Version("0")
        for spec in specifier:
            if spec.operator == ">=":
                minimum_version = max(minimum_version, Version(spec.version))
            elif spec.operator == "==":
                minimum_version = Version(spec.version)
        return minimum_version

    def is_package_version_too_old(self, package: str, specifier: SpecifierSet) -> bool:
        exist_package = self._packages_distributions.get(package)
        if exist_package is None:
            return False
        exist_version = Version(exist_package.version)
        minimum_version = self._get_minimum_version(specifier)
        return minimum_version < exist_version

    async def find_best_compatible_version(self, package: str, specifier: SpecifierSet) -> Version:
        package_info = await self.fetch_package_info(package)
        versions: list[Version] = sorted(map(Version, package_info["releases"].keys()))  # [0.1.0, 0.1.1, 0.2.0, 2.0.0]
        for version in versions:
            if version in specifier:
                return version
        raise ValueError(f"No compatible version found for {package} with specifier {specifier}")

    async def add_dependencies(self, dependencies: Mapping[str, str | None]) -> bool:
        changed = False
        new_dependencies: dict[str, SpecifierSet] = {}
        for package, specifier in dependencies.items():
            specifier = SpecifierSet(specifier or "")
            package_info = self._packages_distributions.get(package)
            if package_info is None:
                new_dependencies[package] = specifier
                changed = True
                continue
            satisfied = self.is_package_satisfied(package, specifier)
            if satisfied:
                continue
            too_old = self.is_package_version_too_old(package, specifier)
            if too_old:
                raise RequiredVersionTooOld(f"Package {package} is too old: {package_info.version}")
            best_version = await self.find_best_compatible_version(package, specifier)
            new_dependencies[package] = SpecifierSet(f"=={best_version}")
            changed = True
        self._dependencies.update(new_dependencies)
        return changed

    def find_packages_distributions(
        self,
    ) -> Mapping[str, importlib.metadata.Distribution]:
        if not self._distributions_change_marked:
            return self._packages_distributions
        self._packages_distributions: Mapping[str, importlib.metadata.Distribution] = {
            dist.name: dist for dist in importlib.metadata.distributions()
        }
        self._distributions_change_marked = False
        return self._packages_distributions

    async def resolve(self) -> Result[ResolveResult, str]:
        packages_distributions = self.find_packages_distributions()
        requirements: dict[str, SpecifierSet] = {}
        update_distributions: dict[str, importlib.metadata.Distribution] = {}
        new_distributions: list[str] = []
        skipped: dict[str, SpecifierSet] = {}
        for dependency, specifier in self._dependencies.items():
            exist_package = packages_distributions.get(dependency)
            if exist_package is None:
                requirements[dependency] = specifier
                new_distributions.append(dependency)
                continue
            installed_version = Version(exist_package.version)
            specifier_set = self._dependencies[dependency]
            if installed_version in specifier_set:
                skipped[dependency] = specifier_set
                continue
            requirements[dependency] = specifier_set
            update_distributions[dependency] = exist_package
        if len(requirements) == 0:
            return Ok(ResolveResult(new_packages=new_distributions, updated_packages=update_distributions))

        result = await self.install_requirements(requirements)
        if is_err(result):
            return Err(result.err)
        self._distributions_change_marked = True
        logger.info(f"New packages: {new_distributions}")
        logger.info(f"Updated packages: {update_distributions}")
        return Ok(ResolveResult(new_packages=new_distributions, updated_packages=update_distributions))

    def solve_dependency_order(self, entrypoints: dict[str, PluginEntry]):
        # sort dependenies by entrypoint.dist.requires
        normalized_map = {normalize_package_name(k): k for k in entrypoints.keys()}
        sorted_packages: list[str] = []
        visited: set[str] = set()
        temp_mark: set[str] = set()

        def visit(package: str):
            if package in visited:
                return
            if package in temp_mark:
                raise Exception(f"Cyclic dependency detected: {' -> '.join(temp_mark)} -> {package}")
            temp_mark.add(package)
            entry_point = entrypoints.get(package)
            if entry_point is not None and entry_point.dist is not None:
                for req in entry_point.dist.requires or []:
                    requirement = Requirement(req)
                    dep_name = normalize_package_name(requirement.name)
                    dep_package = normalized_map.get(dep_name)
                    if dep_package is not None:
                        visit(dep_package)
            temp_mark.remove(package)
            visited.add(package)
            sorted_packages.append(package)

        for pkg in entrypoints.keys():
            visit(pkg)
        return [(pkg, entrypoints[pkg]) for pkg in sorted_packages]


@dataclass(frozen=True, slots=True)
class ResolveResult:
    new_packages: list[str]
    updated_packages: dict[str, importlib.metadata.Distribution]


class PluginLoader:
    def __init__(
        self,
        server: Server,
        resolver: DependencyResolver,
    ) -> None:
        self._server = server
        self.instances: dict[str, PluginInstance] = {}
        self.resolver = resolver

    async def load_plugins(self):
        ordered_entries = self.resolver.solve_dependency_order(PluginEntry.retrieve_plugin_entries())
        for package, entry in ordered_entries:
            if package in self.instances:
                logger.warning(f"Skip loading plugin: {package}({entry})")
                continue
            match PluginInstance.try_load(entry):
                case Ok(instance):
                    self.instances[entry.key] = instance
                case Err(message):
                    logger.warning(f"Failed to load plugin {entry.name}: {message}")

        logger.info(f"Loaded plugins: {', '.join(self.instances.keys())}")

    async def update_plugins(self, resolve_result: ResolveResult):
        new_entries: dict[str, PluginEntry] = {}
        updated_entries: dict[str, PluginEntry] = {}
        for package in resolve_result.new_packages:
            distribution = importlib.metadata.distribution(package)
            for key, entry in PluginEntry.retrieve_from_distribution(distribution).items():
                if key in new_entries:
                    logger.error(f"Duplicate plugin {key} during installation")
                    continue
                new_entries[key] = entry

        for package in resolve_result.updated_packages.keys():
            distribution = importlib.metadata.distribution(package)
            for key, entry in PluginEntry.retrieve_from_distribution(distribution).items():
                if key in updated_entries:
                    logger.error(f"Duplicate plugin {key} during update")
                    continue
                updated_entries[key] = entry

        instances_to_start: list[PluginInstance] = []
        restart_requires: list[str] = []

        for key, entry in self.resolver.solve_dependency_order(new_entries):
            if key in self.instances:
                logger.error(f"Plugin {key} already exists during installation")
                continue
            match PluginInstance.try_load(entry):
                case Err(message):
                    logger.error(f"Error loading plugin {key}: {message}")
                    continue
                case Ok(instance):
                    ...
            self.instances[key] = instance
            ctx = InstallContext(
                server=self._server,
                new_distribution=entry.dist,
            )
            await instance.notify_install(ctx)
            if ctx.restart_required:
                restart_requires.append(key)
            instances_to_start.append(instance)

        for key, entry in self.resolver.solve_dependency_order(updated_entries):
            instance = self.instances.get(key)
            if instance is None:
                continue
            await instance.terminate(self._server)
            await instance.reload()
            ctx = InstallContext(
                server=self._server,
                old_distribution=instance.entry.dist,
                new_distribution=entry.dist,
                old_plugin=instance.plugin,
            )
            await instance.notify_update(ctx)
            await instance.notify_install(ctx)
            if ctx.restart_required:
                restart_requires.append(key)
            instances_to_start.append(instance)

        if restart_requires:
            logger.info(f"Restart required for plugins: {restart_requires}")
            await self._server.restart()
            return

        for instance in instances_to_start:
            await instance.start(self._server)

    async def stop_plugins(self):
        for instance in self.instances.values():
            await instance.terminate(self._server)
        self.instances.clear()



from __future__ import annotations

import asyncio
import importlib
import importlib.metadata
import os
import sys
import threading
import time
from dataclasses import dataclass
from multiprocessing import Process
from types import ModuleType
from typing import TYPE_CHECKING

import psutil
from loguru import logger
from omu.address import Address
from omu.app import App, AppType
from omu.helper import asyncio_error_logger
from omu.network.websocket_connection import WebsocketsTransport
from omu.omu import Omu
from omu.plugin import InstallContext, Plugin, StartContext, StopContext, UninstallContext
from omu.result import Err, Ok, Result
from omu.token import TokenProvider

from omuserver.helper import setup_logger
from omuserver.session import Session

from .loader import PluginEntry
from .session_connection import PluginSessionConnection
from .transport import PluginConnection

if TYPE_CHECKING:
    from omuserver.server import Server


class PluginTokenProvider(TokenProvider):
    def __init__(self, token: str):
        self._token = token

    def get(self, address: Address, app: App) -> str | None:
        return self._token

    def store(self, address: Address, app: App, token: str) -> None:
        raise NotImplementedError


def deep_reload(module: ModuleType) -> None:
    to_reload: list[ModuleType] = [module]
    module_key = module.__name__ + "."
    for key, module in sys.modules.items():
        if key.startswith(module_key):
            to_reload.append(module)
    for module in to_reload:
        try:
            importlib.reload(module)
        except Exception as e:
            logger.opt(exception=e).error(f"Error reloading module {module}")


@dataclass(slots=True)
class PluginInstance:
    plugin: Plugin
    entry: PluginEntry
    module: ModuleType
    process: Process | None = None
    omu: Omu | None = None

    @classmethod
    def try_load(
        cls,
        entry: PluginEntry,
    ) -> Result[PluginInstance, str]:
        package = entry.dist
        stage = "loading"
        try:
            plugin = entry.entry_point.load()
            stage = "validating"
            if not isinstance(plugin, Plugin):
                return Err(f"{plugin} is not a Plugin")
            stage = "importing"
            module = importlib.import_module(entry.entry_point.module)
            return Ok(
                cls(
                    plugin=plugin,
                    entry=entry,
                    module=module,
                )
            )
        except Exception as e:
            logger.opt(exception=e).error(
                f"Error while {stage} plugin {entry.name} from {package.name if package else 'unknown'}"
            )
            return Err(f"Error while {stage} plugin {entry.name}: {e}")

    async def notify_install(self, ctx: InstallContext):
        if self.plugin.on_install is not None:
            await self.plugin.on_install(ctx)

    async def notify_uninstall(self, ctx: UninstallContext):
        if self.plugin.on_uninstall is not None:
            await self.plugin.on_uninstall(ctx)

    async def notify_update(self, ctx: InstallContext):
        if self.plugin.on_update is not None:
            await self.plugin.on_update(ctx)

    async def reload(self):
        deep_reload(self.module)
        new_plugin = self.entry.entry_point.load()
        if not isinstance(new_plugin, Plugin):
            raise ValueError(f"Invalid plugin: {new_plugin} is not a Plugin")
        self.plugin = new_plugin

    async def terminate(self, server: Server):
        if self.process is not None:
            try:
                self.process.terminate()
                self.process.join()
            except AttributeError:
                logger.warning(f"Error terminating plugin {self.entry.name}")
            except Exception as e:
                logger.opt(exception=e).error(f"Error terminating plugin {self.entry.name}")
            self.process = None
        if self.omu is not None:
            await self.omu.stop()
            self.omu = None
        if self.plugin.on_stop is not None:
            await self.plugin.on_stop(StopContext(server=server))

    async def start(self, server: Server):
        stage = "invoking on_start"
        try:
            if self.plugin.on_start is not None:
                await self.plugin.on_start(StartContext(server=server))
            stage = "generating token"
            token = server.security.generate_plugin_token()
            if self.plugin.isolated:
                stage = "starting isolated"
                result = self._start_isolated(server, token)
            else:
                stage = "starting internally"
                result = await self._start_internally(server, token)
            if result.is_err is True:
                logger.warning(f"Plugin start failed: {result.err}")
        except Exception as e:
            logger.opt(exception=e).error(f"Error while {stage} plugin {self.entry.name}")

    async def _start_internally(self, server: Server, token: str) -> Result[..., str]:
        if self.omu:
            return Err(f'Plugin "{self.entry.key}" already started')
        if self.plugin.get_client is None:
            logger.warning(f'Plugin "{self.entry.key}" has no client')
            return Ok(...)
        connection = PluginConnection()
        self.omu = self.plugin.get_client()
        if self.omu.app.type != AppType.PLUGIN:
            return Err(f"Invalid plugin: {self.omu.app} is not a plugin")
        self.omu.network.set_connection(connection)
        self.omu.network.set_token_provider(PluginTokenProvider(token))
        self.omu.set_loop(server.loop)
        self.omu.loop.create_task(self.omu.start(reconnect=False))
        session_connection = PluginSessionConnection(connection)
        session_result = await Session.from_connection(
            server,
            server.packets.packet_mapper,
            session_connection,
        )
        if session_result.is_err is True:
            return Err(f"Creating internal plugin session failed for {self.entry.key}: {session_result.err}")
        session, _ = session_result.value
        server.loop.create_task(server.sessions.process_new(session))
        return Ok(...)

    def _start_isolated(self, server: Server, token: str) -> Result[..., str]:
        pid = os.getpid()
        if self.process:
            return Err(f'Plugin "{self.plugin}" already started')
        process = Process(
            target=run_plugin_isolated,
            args=(
                self.entry.entry_point,
                server.address,
                token,
                pid,
            ),
            name=f"Plugin {self.entry.entry_point.value}",
            daemon=True,
        )
        process.start()
        self.process = process
        return Ok(...)


def run_plugin_isolated(
    entry_point: importlib.metadata.EntryPoint,
    address: Address,
    token: str,
    pid: int,
) -> None:
    def _watch_parent_process():
        while True:
            if not psutil.pid_exists(pid):
                logger.info(f"Parent process {pid} is dead, stopping plugin")
                exit(0)
            time.sleep(1)

    threading.Thread(target=_watch_parent_process, daemon=True).start()

    package = entry_point.dist
    stage = "loading"
    try:
        plugin = entry_point.load()
        stage = "validating"
        if not isinstance(plugin, Plugin):
            raise ValueError(f"Invalid plugin: {plugin} is not a Plugin")
        stage = "starting"
        if plugin.get_client is None:
            raise ValueError(f"Invalid plugin: {plugin} has no client")
        client = plugin.get_client()
        if client.app.type != AppType.PLUGIN:
            raise ValueError(f"Invalid plugin: {client.app} is not a plugin")
        stage = "setting up"
        setup_logger(name=client.app.id.get_sanitized_key())
        logger.info(f"Starting plugin {client.app.id} {client.app.version}")
        transport = WebsocketsTransport(address)
        client.network.set_transport(transport)
        client.network.set_token_provider(PluginTokenProvider(token))
        loop = asyncio.new_event_loop()
        loop.set_exception_handler(asyncio_error_logger)

        def stop_plugin():
            logger.info(f"Stopping plugin {client.app.id}")
            loop.stop()
            exit(0)

        client.network.event.disconnected += stop_plugin
        stage = "running"
        client.run(loop=loop, reconnect=False)
        loop.run_forever()
    except Exception as e:
        logger.opt(exception=e).error(
            f"Error while {stage} plugin {entry_point.name} from {package.name if package else 'unknown'}"
        )
        return None



from omu.api.permission.permission import PermissionType
from omu.api.session.extension import (
    GENERATE_TOKEN_PERMISSION_ID,
    REMOTE_APP_REQUEST_PERMISSION_ID,
    SESSIONS_READ_PERMISSION_ID,
)

SESSIONS_READ_PERMISSION = PermissionType(
    id=SESSIONS_READ_PERMISSION_ID,
    metadata={
        "level": "low",
        "name": {
            "ja": "接続中のアプリを取得",
            "en": "Get Running Apps",
        },
        "note": {
            "ja": "どのアプリが接続されているかを確認するために使われます",
            "en": "Used to get a list of apps connected to the server",
        },
    },
)

REMOTE_APP_REQUEST_PERMISSION = PermissionType(
    REMOTE_APP_REQUEST_PERMISSION_ID,
    metadata={
        "level": "high",
        "name": {
            "ja": "遠隔アプリを要求",
            "en": "Request Remote App",
        },
        "note": {
            "ja": "ネットワークを経由して操作を可能にするために使われます",
            "en": "Used to enable control over the network",
        },
    },
)

GENERATE_TOKEN_PERMISSION = PermissionType(
    GENERATE_TOKEN_PERMISSION_ID,
    metadata={
        "level": "low",
        "name": {
            "ja": "認証トークンを生成",
            "en": "Generate Auth Token",
        },
        "note": {
            "ja": "OBSにアプリを追加したり、複数画面でアプリを起動するために使われます",
            "en": "Used to add apps to OBS or launch apps on multiple screens",
        },
    },
)



from __future__ import annotations

import ipaddress
import secrets
import socket
import sqlite3
from pathlib import Path
from typing import TYPE_CHECKING, Any

import aiohttp
from aiohttp import web
from loguru import logger
from omu.api.asset.extension import (
    ASSET_DELETE_ENDPOINT,
    ASSET_DOWNLOAD_ENDPOINT,
    ASSET_DOWNLOAD_MANY_ENDPOINT,
    ASSET_GENERATE_TOKEN_ENDPOINT,
    ASSET_UPLOAD_ENDPOINT,
    ASSET_UPLOAD_MANY_ENDPOINT,
    Asset,
    GenerateAssetTokenResponse,
)
from omu.errors import PermissionDenied
from omu.identifier import Identifier
from yarl import URL

from omuserver.helper import safe_path_join
from omuserver.session import Session

from .permissions import (
    ASSET_PERMISSION,
)

if TYPE_CHECKING:
    from omuserver.server import Server


class AssetIndex:
    def __init__(self, index_db: Path) -> None:
        self.index_db = sqlite3.connect(index_db)
        self.init_db()
        self._index_cache: dict[Identifier, Path] = {}

    def init_db(self):
        cursor = self.index_db.cursor()
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS assets (
            id TEXT PRIMARY KEY,
            path TEXT,
            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            updated_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            download_count INTEGER DEFAULT 0
            )
            """
        )
        self.index_db.commit()

    def insert(self, identifier: Identifier, path: Path) -> None:
        if identifier in self._index_cache:
            return
        self._index_cache[identifier] = path
        cursor = self.index_db.cursor()
        cursor.execute(
            """
            INSERT OR IGNORE INTO assets (id, path)
            VALUES (?, ?)
            """,
            (identifier.key(), path.as_posix()),
        )
        self.index_db.commit()

    def delete(self, identifier: Identifier) -> None:
        cursor = self.index_db.cursor()
        cursor.execute(
            """
            DELETE FROM assets
            WHERE id = ?
            """,
            (identifier.key(),),
        )
        self.index_db.commit()
        if identifier in self._index_cache:
            del self._index_cache[identifier]

    def lookup_asset_path(self, identifier: Identifier) -> Path | None:
        if identifier in self._index_cache:
            return self._index_cache[identifier]
        cursor = self.index_db.cursor()
        cursor.execute(
            """
            SELECT path
            FROM assets
            WHERE id = ?
            """,
            (identifier.key(),),
        )
        row = cursor.fetchone()
        if row is None:
            return None
        path = Path(row[0])
        self._index_cache[identifier] = path
        return path

    def record_download(self, identifier: Identifier) -> None:
        cursor = self.index_db.cursor()
        cursor.execute(
            """
            UPDATE assets
            SET download_count = download_count + 1, updated_at = CURRENT_TIMESTAMP
            WHERE id = ?
            """,
            (identifier.key(),),
        )
        self.index_db.commit()

    def record_upload(self, identifier: Identifier) -> None:
        cursor = self.index_db.cursor()
        cursor.execute(
            """
            UPDATE assets
            SET updated_at = CURRENT_TIMESTAMP
            WHERE id = ?
            """,
            (identifier.key(),),
        )
        self.index_db.commit()


class AssetExtension:
    def __init__(self, server: Server) -> None:
        server.security.register_permission(ASSET_PERMISSION)
        server.endpoints.bind(ASSET_GENERATE_TOKEN_ENDPOINT, self.handle_generate_token)
        server.endpoints.bind(ASSET_UPLOAD_ENDPOINT, self.handle_upload)
        server.endpoints.bind(ASSET_UPLOAD_MANY_ENDPOINT, self.handle_upload_many)
        server.endpoints.bind(ASSET_DOWNLOAD_ENDPOINT, self.handle_download)
        server.endpoints.bind(ASSET_DOWNLOAD_MANY_ENDPOINT, self.handle_download_many)
        server.endpoints.bind(ASSET_DELETE_ENDPOINT, self.handle_delete)
        self.session_asset_tokens: dict[Identifier, str] = {}
        self.asset_tokens: set[str] = set()
        self.server = server
        self._path = server.directories.assets
        self.index = AssetIndex(self.server.directories.assets / "index.sqlite")
        server.network.route_get("/proxy", self._handle_proxy)
        server.network.route_get("/asset", self._handle_assets)

    def _verify_asset_token(self, request: web.Request):
        token = request.query.get("asset_token")
        if token is None:
            raise Exception("asset_token is not set")
        if token not in self.asset_tokens:
            raise Exception("Invalid asset_token was passed")

    async def _handle_proxy(self, request: web.Request) -> web.StreamResponse:
        self._verify_asset_token(request)
        url = request.query.get("url")
        if not url:
            return web.Response(status=400)
        url = URL(url)
        if url.host is None:
            raise Exception("Invalid url: host is not specified")
        ip_address = socket.gethostbyname(url.host)
        address = ipaddress.ip_address(ip_address)
        if address.is_private:
            return web.Response(status=403)
        no_cache = bool(request.query.get("no_cache"))
        try:
            async with self.server.client.get(
                url,
            ) as resp:
                headers = {
                    "Cache-Control": "no-cache" if no_cache else "max-age=3600",
                    "Content-Type": resp.content_type,
                    "Access-Control-Allow-Origin": "*",
                }
                response = web.StreamResponse(status=resp.status, headers=headers)
                await response.prepare(request)
                async for chunk in resp.content.iter_any():
                    await response.write(chunk)
                return response
        except TimeoutError:
            return web.Response(status=504)
        except aiohttp.ClientConnectionResetError:
            return web.Response(status=502)
        except aiohttp.ClientResponseError as e:
            return web.Response(status=e.status, text=e.message)
        except Exception:
            logger.error("Failed to proxy request")
            return web.Response(status=500)

    async def _handle_assets(self, request: web.Request) -> web.StreamResponse:
        self._verify_asset_token(request)
        id = request.query.get("id")
        if not id:
            return web.Response(status=400)
        identifier = Identifier.from_key(id)
        path = identifier.get_sanitized_path()
        try:
            path = safe_path_join(self.server.directories.assets, path)

            if not path.exists():
                return web.Response(status=404)
            return web.FileResponse(path)
        except Exception as e:
            logger.error(e)
            return web.Response(status=500)

    async def store(self, file: Asset) -> Identifier:
        path = file.id.get_sanitized_path()
        file_path = safe_path_join(self._path, path)
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.write_bytes(file.buffer)
        self.index.insert(file.id, path)
        self.index.record_upload(file.id)
        return file.id

    async def retrieve(self, identifier: Identifier) -> Asset | None:
        path = identifier.get_sanitized_path()
        file_path = safe_path_join(self._path, path)
        if not file_path.exists():
            return None
        self.index.insert(identifier, path)
        self.index.record_download(identifier)
        return Asset(identifier, file_path.read_bytes())

    async def handle_generate_token(self, session: Session, args: Any) -> GenerateAssetTokenResponse:
        existing = self.session_asset_tokens.get(session.id)
        if existing:
            return {"token": existing}
        token = self.session_asset_tokens[session.id] = secrets.token_urlsafe()
        self.asset_tokens.add(token)
        return {"token": token}

    async def handle_upload(self, session: Session, file: Asset) -> Identifier:
        if not session.is_app_id(file.id):
            raise PermissionDenied(f"App {session.app.id=} not allowed to {file.id}")
        identifier = await self.store(file)
        return identifier

    async def handle_upload_many(self, session: Session, files: list[Asset]) -> list[Identifier]:
        asset_ids: list[Identifier] = []
        for file in files:
            if not session.is_app_id(file.id):
                raise PermissionDenied(f"App {session.app.id=} not allowed to {file.id}")
            id = await self.store(file)
            asset_ids.append(id)
        return asset_ids

    async def handle_download(self, session: Session, id: Identifier) -> Asset:
        if not session.is_app_id(id):
            raise PermissionDenied(f"App {session.app.id=} not allowed to {id}")
        asset = await self.retrieve(id)
        if asset is None:
            raise Exception(f"Asset {id} not found")
        return asset

    async def handle_download_many(self, session: Session, identifiers: list[Identifier]) -> list[Asset]:
        added_files: list[Asset] = []
        for id in identifiers:
            if not session.is_app_id(id):
                raise PermissionDenied(f"App {session.app.id=} not allowed to {id}")
            asset = await self.retrieve(id)
            if asset is None:
                raise Exception(f"Asset {id} not found")
            added_files.append(asset)
        return added_files

    async def handle_delete(self, session: Session, id: Identifier) -> None:
        if not session.is_app_id(id):
            raise PermissionDenied(f"App {session.app.id=} not allowed to {id}")
        path = self.index.lookup_asset_path(id)
        if path is None:
            raise Exception(f"Asset {id} not found")
        file_path = safe_path_join(self._path, path)
        file_path.unlink()
        self.index.delete(id)



from __future__ import annotations

import signal
import socket
import sys
import tkinter
from pathlib import Path
from tkinter import messagebox
from typing import TYPE_CHECKING

import aiohttp
import aiohttp.typedefs
import psutil
from aiohttp import web
from loguru import logger
from omu import Identifier
from omu.address import get_lan_ip
from omu.app import AppType
from omu.errors import DisconnectReason, InvalidOrigin
from omu.event_emitter import EventEmitter
from omu.helper import Coro
from omu.network.packet import PACKET_TYPES, PacketType
from omu.network.packet.packet_types import DisconnectPacket, DisconnectType
from omu.result import Err, Ok, Result, is_err
from psutil import Process
from yarl import URL

from omuserver.helper import find_processes_by_executable, find_processes_by_port, get_parent_pids
from omuserver.network.aiohttp_connection import WebsocketsConnection
from omuserver.network.packet_dispatcher import ServerPacketDispatcher
from omuserver.session import Session

if TYPE_CHECKING:
    from omuserver.server import Server


@web.middleware
async def cors_middleware(request: web.Request, handler: aiohttp.typedefs.Handler) -> web.StreamResponse:
    if request.method == "OPTIONS":
        return web.Response(
            status=200,
            headers={
                "Access-Control-Allow-Origin": "*",
                "Access-Control-Allow-Headers": "*",
                "Access-Control-Allow-Methods": "*",
            },
        )
    response = await handler(request)
    response.headers["Access-Control-Allow-Origin"] = request.headers.get("Origin", "*")
    response.headers["Access-Control-Allow-Headers"] = "*"
    response.headers["Access-Control-Allow-Methods"] = "*"
    return response


class Network:
    def __init__(self, server: Server, packet_dispatcher: ServerPacketDispatcher) -> None:
        self.server = server
        self._packet_dispatcher = packet_dispatcher
        self._app = web.Application(middlewares=[cors_middleware])
        self._runner: web.AppRunner | None = None
        self._app.router.add_get("/ws", self.websocket_handler)
        self.register_packet(
            PACKET_TYPES.CONNECT,
            PACKET_TYPES.READY,
            PACKET_TYPES.DISCONNECT,
            PACKET_TYPES.ENCRYPTED_PACKET,
        )
        self.add_packet_handler(PACKET_TYPES.READY, self._handle_ready)
        self.add_packet_handler(PACKET_TYPES.DISCONNECT, self._handle_disconnection)
        self.local_ip = get_lan_ip()

    async def stop(self) -> None:
        if self._runner is None:
            raise ValueError("Server not started")
        await self._app.shutdown()
        await self._app.cleanup()

    async def _handle_ready(self, session: Session, packet: None) -> None:
        await session.process_ready_tasks()
        if session.closed:
            return
        await session.send(PACKET_TYPES.READY, None)
        parts = [session.app.key()]
        if session.app.version is not None:
            parts.append(f"v{session.app.version}")
        logger.info(f"Ready: {' '.join(parts)}")

    async def _handle_disconnection(self, session: Session, packet: DisconnectPacket) -> None:
        await session.disconnect(DisconnectType.CLOSE, packet.message)

    def register_packet(self, *packet_types: PacketType) -> None:
        self._packet_dispatcher.register(*packet_types)

    def add_packet_handler[T](
        self,
        packet_type: PacketType[T],
        coro: Coro[[Session, T], None],
    ) -> None:
        self._packet_dispatcher.bind(packet_type, coro)

    def route_get(self, path: str, handle: Coro[[web.Request], web.StreamResponse]) -> None:
        self._app.router.add_get(path, handle)

    def route_post(self, path: str, handle: Coro[[web.Request], web.StreamResponse]) -> None:
        self._app.router.add_post(path, handle)

    def _verify_remote_ip(self, request: web.Request, session: Session) -> Result[None, DisconnectReason]:
        if request.remote not in {self.local_ip, "127.0.0.1"}:
            logger.warning(f"Invalid remote ip {request.remote} for {session.app}")
            return Err(InvalidOrigin("Invalid remote ip (see logs)"))
        return Ok(None)

    def get_mapped_host(self, host: str) -> str:
        hosts = {
            **self.server.server.trusted_hosts.get(),
            **self.server.config.extra_trusted_hosts,
        }
        return hosts.get(host) or host

    def get_mapped_url(self, url: URL) -> URL:
        parsed = URL(url)
        authority = self.get_mapped_host(parsed.authority)
        return URL.build(
            scheme=url.scheme,
            authority=authority,
            path=url.path,
            query_string=url.query_string,
            fragment=url.fragment,
        )

    async def _verify_origin(self, request: web.Request, session: Session) -> Result[None, DisconnectReason]:
        origin = request.headers.get("Origin")
        if origin is None:
            return Ok(None)
        origin_url = self.get_mapped_url(URL(origin))
        origin_namespace = Identifier.namespace_from_url(origin_url)
        session_namespace = session.app.id.namespace
        if origin_namespace == session_namespace:
            return Ok(None)
        return Err(InvalidOrigin(f"Invalid origin: {origin_namespace} != {session_namespace}"))

    async def _verify(self, session: Session, request: web.Request) -> Result[None, DisconnectReason]:
        if session.kind == AppType.REMOTE:
            return self.server.security.verify_frame_token(request)
        ip_verified = self._verify_remote_ip(request, session)
        if session.kind in {AppType.DASHBOARD, AppType.PLUGIN}:
            return ip_verified
        origin_verified = await self._verify_origin(request, session)
        return ip_verified and origin_verified

    async def websocket_handler(self, request: web.Request) -> web.StreamResponse:
        ws = web.WebSocketResponse(max_msg_size=0)
        await ws.prepare(request)
        connection = WebsocketsConnection(ws)
        session_result = await Session.from_connection(
            self.server,
            self._packet_dispatcher.packet_mapper,
            connection,
        )
        if session_result.is_err is True:
            logger.warning(f"Connection failed: {session_result.err}")
            return web.Response(status=403)
        session, new_token = session_result.value

        verify_result = await self._verify(session, request)
        if is_err(verify_result):
            logger.warning(f"Verification failed for {session.app}: {verify_result.err}")
            await session.disconnect(verify_result.err.type, verify_result.err.message)
            return web.Response(status=403)

        if session.kind not in {AppType.PLUGIN, AppType.DASHBOARD}:
            await session.send(PACKET_TYPES.TOKEN, new_token)

        await self.server.sessions.process_new(session)
        return ws

    def is_port_free(self) -> bool:
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(("0.0.0.0", self.server.address.port))
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(("127.0.0.1", self.server.address.port))
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind((self.server.address.host, self.server.address.port))
            return True
        except OSError:
            return False

    def _is_server_process(self, process: Process) -> bool:
        args = process.cmdline()
        if len(args) >= 2 and "python.debugpy" in args[1]:
            return False
        return True

    def ensure_port_availability(self) -> Result[None, tuple[str, list[Process]]]:
        if self.is_port_free():
            return Ok(None)
        executable = Path(sys.executable)
        port = self.server.address.port
        found_processes: set[Process] = set()
        found_processes.update(find_processes_by_port(port))
        logger.info(f"No processes found using port {port}")
        found_processes.update(find_processes_by_executable(executable))
        ignore_pids = list(get_parent_pids())
        found_processes = {p for p in found_processes if (p.pid not in ignore_pids) and (self._is_server_process(p))}
        if len(found_processes) == 0:
            return Err((f"Port {port} already in use by unknown process", []))
        if len(found_processes) > 1:
            processes = " ".join(f"{p.name()} ({p.pid=})" for p in found_processes)
            return Err(
                (
                    f"Port {port} already in use by multiple processes: {processes}",
                    list(found_processes),
                )
            )
        process = found_processes.pop()
        port = self.server.address.port
        name = process.name()
        pid = process.pid
        parents = process.parents()
        msg = f"Port {port} already in use by {' -> '.join(f'{p.name()}({p.pid})' for p in parents)} -> {name} ({pid=})"
        return Err((msg, list(found_processes)))

    def terminate_process_due_to_port(self, processes: list[Process]):
        root = tkinter.Tk()
        root.withdraw()
        root.attributes("-topmost", True)

        def wait_for_process_to_end():
            if any(process.is_running() for process in processes):
                root.after(200, wait_for_process_to_end)
            else:
                root.destroy()

        root.after(200, wait_for_process_to_end)

        process_names = ", ".join(process.name() for process in processes)
        message = f"OMUAPPSの使用するためには以下のアプリを閉じる必要があります\n{process_names}"
        res = messagebox.Message(
            root,
            title="OMUAPPS API",
            message=message,
            icon=messagebox.WARNING,
            type=messagebox.YESNO,
        ).show()
        if not res or res == messagebox.YES:
            ignore_pids = list(get_parent_pids())
            for process in processes:
                try:
                    if process.pid in ignore_pids:
                        continue
                    logger.info(f"Killing process {process.pid} ({process.name()}) running at {process.cwd()}")
                    if not process.is_running():
                        continue
                    process.send_signal(signal.SIGTERM)
                except psutil.NoSuchProcess:
                    logger.warning(f"Process {process.pid} not found")
                except psutil.AccessDenied:
                    logger.warning(f"Access denied to process {process.pid}")
                except Exception:
                    logger.exception(f"Failed to kill process {process.pid}")
        elif res == messagebox.NO:
            return Err("User cancelled shutdown")

    def notify_system_idle_process_conflict(self):
        root = tkinter.Tk()
        root.withdraw()
        root.attributes("-topmost", True)
        messagebox.Message(
            root,
            title="OMUAPPS API",
            message="起動に失敗しました再度お試しください。これにはPCの再起動が必要な場合があります",
            icon=messagebox.WARNING,
            type=messagebox.OK,
        ).show()

    async def start(self) -> None:
        if self._runner is not None:
            raise ValueError("Server already started")
        match self.ensure_port_availability():
            case Err((msg, processes)):
                logger.error(msg)
                if len(processes) == 0:
                    self.notify_system_idle_process_conflict()
                self.terminate_process_due_to_port(processes)

        runner = web.AppRunner(self._app)
        self._runner = runner
        await runner.setup()
        site = web.TCPSite(
            runner,
            host="0.0.0.0",
            port=self.server.address.port,
        )
        await site.start()


class NetworkEvents:
    def __init__(self) -> None:
        self.connected = EventEmitter[Session]()
        self.disconnected = EventEmitter[Session]()



from dataclasses import dataclass, field

from omu.address import Address

from omuserver.directories import Directories


@dataclass(slots=True)
class Config:
    address: Address = Address.default()
    debug: bool = False
    extra_trusted_hosts: dict[str, str] = field(default_factory=dict)
    directories: Directories = field(default_factory=Directories.default)
    dashboard_token: str | None = None
    index_url: str | None = None



from __future__ import annotations

import abc
import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING

from loguru import logger
from omu import App
from omu.app import AppType
from omu.errors import DisconnectReason
from omu.event_emitter import EventEmitter
from omu.helper import Coro
from omu.identifier import Identifier
from omu.network.connection import InvalidPacket, ReceiveError
from omu.network.encryption import AES, Decryptor
from omu.network.packet import PACKET_TYPES, Packet, PacketType
from omu.network.packet.packet_types import DisconnectPacket, DisconnectType, ServerMeta
from omu.network.packet_mapper import PacketMapper
from omu.result import Err, Ok, Result

from omuserver.brand import BRAND
from omuserver.security import InputToken
from omuserver.version import VERSION

if TYPE_CHECKING:
    from omuserver.security import PermissionHandle
    from omuserver.server import Server


class SessionConnection(abc.ABC):
    @abc.abstractmethod
    async def send(self, packet: Packet, packet_mapper: PacketMapper) -> Result[..., str]: ...

    @abc.abstractmethod
    async def receive(self, packet_mapper: PacketMapper) -> Result[Packet, ReceiveError]: ...

    async def receive_as[T](self, packet_mapper: PacketMapper, packet_type: PacketType[T]) -> Result[T, ReceiveError]:
        packet = await self.receive(packet_mapper)
        if packet.is_err is True:
            return Err(packet.err)
        packet = packet.value
        if packet.type != packet_type:
            return Err(InvalidPacket(f"Expected {packet_type.id} but got {packet.type}"))
        return Ok(packet.data)

    @abc.abstractmethod
    async def close(self) -> None: ...

    @property
    @abc.abstractmethod
    def closed(self) -> bool: ...


class SessionEvents:
    def __init__(self) -> None:
        self.disconnected = EventEmitter[Session](catch_errors=True)
        self.ready = EventEmitter[Session]()


class TaskPriority:
    AFTER_CONNECTED = 0
    AFTER_PLUGIN = 200
    AFTER_SESSION = 300
    AFTER_PERMITTED = 400


@dataclass(frozen=True, slots=True)
class SessionTask:
    session: Session
    coro: Coro[[], None]
    name: str
    detail: str | None = None
    priority: int = 0

    def __repr__(self) -> str:
        return f"SessionTask(name={self.name}, detail={self.detail}, priority={self.priority})"

    def __str__(self) -> str:
        return self.__repr__()


class Session:
    def __init__(
        self,
        server: Server,
        packet_mapper: PacketMapper,
        app: App,
        permission_handle: PermissionHandle,
        kind: AppType,
        connection: SessionConnection,
        aes: AES | None = None,
    ) -> None:
        self.server = server
        self.packet_mapper = packet_mapper
        self.app = app
        self.id = app.id
        self.permissions = permission_handle
        self.kind = kind
        self.connection = connection
        self.event = SessionEvents()
        self.ready_tasks: list[SessionTask] = []
        self.ready_waiters: list[asyncio.Future[None]] = []
        self.ready: bool = False
        self.aes = aes

    @classmethod
    async def from_connection(
        cls,
        server: Server,
        packet_mapper: PacketMapper,
        connection: SessionConnection,
    ) -> Result[tuple[Session, InputToken], str]:
        decryptor = Decryptor.new()
        meta: ServerMeta = {
            "protocol": {
                "version": VERSION,
                "brand": BRAND,
            },
            "encryption": {
                "kind": "v1",
                "rsa": decryptor.to_request(),
            },
            "hash": server.address.hash,
        }
        await connection.send(PACKET_TYPES.SERVER_META.new(meta), packet_mapper)
        received = await connection.receive_as(packet_mapper, PACKET_TYPES.CONNECT)
        if received.is_err is True:
            await connection.send(
                PACKET_TYPES.DISCONNECT.new(DisconnectPacket(DisconnectType.INVALID_PACKET, received.err.message)),
                packet_mapper,
            )
            await connection.close()
            return Err(f"Invalid packet received while connecting: {received.err}")
        else:
            packet = received.value
        aes: AES | None = None
        token = packet.token
        if packet.encryption:
            aes = AES.deserialize(packet.encryption["aes"], decryptor)
            if token:
                token = decryptor.decrypt_string(token)
        verify_result = await server.security.verify_app(packet.app, InputToken(token))
        if verify_result.is_err is True:
            await connection.send(
                PACKET_TYPES.DISCONNECT.new(DisconnectPacket(DisconnectType.INVALID_TOKEN, verify_result.err)),
                packet_mapper,
            )
            await connection.close()
            return Err(f"Invalid token for {packet.app}: {verify_result.err}")
        permission_handle, new_token = verify_result.value
        session = Session(
            server=server,
            packet_mapper=packet_mapper,
            app=packet.app,
            permission_handle=permission_handle,
            kind=packet.app.type or AppType.APP,
            connection=connection,
            aes=aes,
        )
        return Ok((session, new_token))

    @property
    def closed(self) -> bool:
        return self.connection.closed

    async def disconnect(self, disconnect_type: DisconnectType, message: str | None = None) -> None:
        if not self.connection.closed:
            await self.send(PACKET_TYPES.DISCONNECT, DisconnectPacket(disconnect_type, message))
        await self.connection.close()
        await self.event.disconnected.emit(self)

    async def listen(self) -> None:
        while not self.connection.closed:
            received = await self.connection.receive(self.packet_mapper)
            if received.is_err is True:
                await self.disconnect(DisconnectType.INVALID_PACKET, received.err.message)
                return
            asyncio.create_task(self.dispatch_packet(received.value))

    async def dispatch_packet(self, packet: Packet) -> None:
        try:
            if self.aes:
                packet = self.packet_mapper.deserialize(self.aes.decrypt(packet))
            await self.server.packets.process_packet(self, packet)
        except DisconnectReason as reason:
            logger.opt(exception=reason).error("Disconnecting session")
            await self.disconnect(reason.type, reason.message)

    async def send[T](self, packet_type: PacketType[T], data: T) -> None:
        packet = Packet(packet_type, data)
        if self.aes:
            packet = self.aes.encrypt(self.packet_mapper.serialize(packet))
        result = await self.connection.send(packet, self.packet_mapper)
        result.unwrap()

    def add_task(
        self,
        coro: Coro[[], None],
        name: str,
        detail: str | None = None,
        priority: int = 0,
    ) -> SessionTask:
        if self.ready:
            raise RuntimeError("Session is already ready")

        task = SessionTask(
            session=self,
            coro=coro,
            name=name,
            detail=detail,
            priority=priority,
        )
        self.ready_tasks.append(task)
        return task

    async def wait_ready(self) -> None:
        if self.ready:
            return
        waiter = asyncio.get_running_loop().create_future()
        self.ready_waiters.append(waiter)
        await waiter

    async def process_ready_tasks(self) -> None:
        if self.ready:
            raise RuntimeError("Session is already ready")
        self.ready = True
        self.ready_tasks.sort(key=lambda t: t.priority)
        logger.debug(f"Session {self.app.key()} is ready, processing {self.ready_tasks} ready tasks")
        for task in self.ready_tasks:
            try:
                logger.debug(f"Processing ready task {task} for session {self.app.key()}: {task.detail}")
                await task.coro()
            except DisconnectReason as e:
                logger.opt(exception=e).error(
                    f"Disconnecting session while processing ready task {task.name}: {task.detail}"
                )
                await self.disconnect(e.type, e.message)
                return
            except Exception as e:
                logger.opt(exception=e).error(f"Error while processing ready task {task.name}: {task.detail}")
                await self.disconnect(DisconnectType.INTERNAL_ERROR, "Error while processing ready tasks")
                return
        self.ready_tasks.clear()
        for waiter in self.ready_waiters:
            waiter.set_result(None)
        await self.event.ready.emit(self)
        self.ready = True

    def is_app_id(self, id: Identifier) -> bool:
        return self.app.id.is_namepath_equal(id, max_depth=1)

    def __repr__(self) -> str:
        return f"Session({self.app.key()}, kind={self.kind}, ready={self.ready})"



from __future__ import annotations

import asyncio
import time
from typing import TYPE_CHECKING

from loguru import logger
from omu.api.plugin import PackageInfo
from omu.api.plugin.extension import (
    PLUGIN_ALLOWED_PACKAGE_TABLE,
    PLUGIN_RELOAD_ENDPOINT_TYPE,
    PLUGIN_REQUIRE_PACKET,
    ReloadOptions,
    ReloadResult,
)
from omu.app import AppType
from omu.errors import PermissionDenied
from omu.network.packet.packet_types import DisconnectType
from omu.plugin import UninstallContext
from omu.result import Err
from packaging.specifiers import SpecifierSet

from omuserver.session import Session

from .loader import DependencyResolver, PluginLoader
from .permissions import PLUGIN_MANAGE_PERMISSION, PLUGIN_READ_PERMISSION

if TYPE_CHECKING:
    from omuserver.server import Server
RESTART = True


class PluginExtension:
    def __init__(self, server: Server) -> None:
        self.allowed_packages = server.tables.register(PLUGIN_ALLOWED_PACKAGE_TABLE)
        server.security.register_permission(PLUGIN_MANAGE_PERMISSION, PLUGIN_READ_PERMISSION)
        server.packets.register(PLUGIN_REQUIRE_PACKET)
        server.tables.register(PLUGIN_ALLOWED_PACKAGE_TABLE)
        server.packets.bind(PLUGIN_REQUIRE_PACKET, self.handle_require)
        server.endpoints.bind(PLUGIN_RELOAD_ENDPOINT_TYPE, self.handle_reload)
        server.event.start += self.on_network_start
        server.event.stop += self.on_stop
        self.server = server
        self.request_id = 0
        self.lock = asyncio.Lock()
        self.dependency_resolver = DependencyResolver(server)
        self.loader = PluginLoader(server, self.dependency_resolver)

    async def on_network_start(self) -> None:
        await self.loader.load_plugins()

        for instance in self.loader.instances.values():
            await instance.start(self.server)

    async def on_stop(self) -> None:
        try:
            await self.loader.stop_plugins()
        except Exception as e:
            logger.opt(exception=e).error("Error stopping plugins")

    def _get_next_request_id(self) -> str:
        self.request_id += 1
        return f"{self.request_id}-{time.time_ns()}"

    async def open_request_plugin_dialog(self, session: Session, packages: dict[str, str | None]) -> None:
        to_request: list[PackageInfo] = []
        for package in packages.keys():
            package_info = await self.dependency_resolver.get_installed_package_info(package)
            if package_info is None:
                package_info = await self.dependency_resolver.fetch_package_info(package)
                to_request.append(package_info)
                continue
            await self.allowed_packages.add(package_info)
        if len(to_request) == 0:
            return
        accepted = await self.server.dashboard.request_plugins(app=session.app, packages=to_request)
        if not accepted:
            raise PermissionDenied("Plugin request was denied by the dashboard")

    async def handle_require(self, session: Session, requirements: dict[str, str | None]) -> None:
        if not requirements:
            return

        match self.dependency_resolver.is_requirements_satisfied(
            {k: SpecifierSet(v) if v else None for k, v in requirements.items()}
        ):
            case Err(err):
                await session.disconnect(
                    DisconnectType.INVALID_VERSION, f"Required plugin versions are not installed: {err}"
                )
                return

        if session.kind == AppType.REMOTE:
            await session.disconnect(DisconnectType.PERMISSION_DENIED, "Remote applications cannot require plugins")
            return

    async def handle_reload(self, session: Session, options: ReloadOptions) -> ReloadResult:
        instances = self.loader.instances

        if options.get("packages") is not None:
            filters = options["packages"] or []
            instances = {name: version for name, version in instances.items() if name in filters}

        for instance in instances.values():
            await instance.terminate(self.server)
            await instance.reload()
            await instance.start(self.server)

        return {"packages": {}}

    async def uninstall(self):
        await self.loader.load_plugins()

        for instance in self.loader.instances.values():
            ctx = UninstallContext(
                self.server,
                instance.entry.dist,
                plugin=instance.plugin,
            )
            await instance.notify_uninstall(ctx)



from __future__ import annotations

import time
from typing import TYPE_CHECKING

from omu.api.permission import PermissionType
from omu.api.permission.extension import (
    PERMISSION_GRANT_PACKET,
    PERMISSION_REGISTER_PACKET,
    PERMISSION_REQUEST_ENDPOINT,
    PERMISSION_REQUIRE_PACKET,
)
from omu.app import AppType
from omu.errors import PermissionDenied
from omu.identifier import Identifier

from omuserver.session import Session, TaskPriority

if TYPE_CHECKING:
    from omuserver.server import Server


class PermissionExtension:
    def __init__(self, server: Server) -> None:
        server.packets.register(PERMISSION_REGISTER_PACKET, PERMISSION_REQUIRE_PACKET, PERMISSION_GRANT_PACKET)
        server.packets.bind(PERMISSION_REGISTER_PACKET, self.handle_register)
        server.packets.bind(PERMISSION_REQUIRE_PACKET, self.handle_require)
        server.endpoints.bind(PERMISSION_REQUEST_ENDPOINT, self.handle_request)
        self.server = server
        self.request_key = 0

    async def handle_register(self, session: Session, permission_types: list[PermissionType]) -> None:
        for perm in permission_types:
            if not perm.id.is_subpath_of(session.app.id):
                msg = f"Permission identifier {perm.id} " f"is not a subpart of app identifier {session.app.id}"
                raise ValueError(msg)
        self.server.security.register_permission(*permission_types, overwrite=True)

    async def handle_require(self, session: Session, permission_ids: list[Identifier]):
        if session.ready:
            raise ValueError("Session is already ready")

        async def task():
            permissions = list(
                filter(None, (self.server.security.permissions.get(perm_id) for perm_id in permission_ids))
            )
            missing_permissions = list(filter(lambda id: id not in self.server.security.permissions, permission_ids))
            if missing_permissions:
                raise PermissionDenied(f"Requested permissions do not exist: {missing_permissions}")
            if session.permissions.has_all(permission_ids):
                await session.send(PERMISSION_GRANT_PACKET, permissions)
                return
            if session.kind == AppType.REMOTE:
                raise PermissionDenied("Remote apps cannot request permissions")
            if session.kind in {AppType.PLUGIN, AppType.DASHBOARD}:
                return
            request_id = self._get_next_request_key()
            accepted = await self.server.dashboard.request_permissions(app=session.app, permissions=permissions)
            if not accepted:
                msg = f"Permission request denied (id={request_id})"
                raise PermissionDenied(msg)
            session.permissions.grant_all([p.id for p in permissions])
            if not session.closed:
                await session.send(PERMISSION_GRANT_PACKET, permissions)

        session.add_task(
            task,
            name="permission_require",
            detail=f"Requiring permissions: {', '.join(map(Identifier.key, permission_ids))}",
            priority=TaskPriority.AFTER_SESSION,
        )

    async def handle_request(self, session: Session, permission_identifiers: list[Identifier]): ...

    def _get_next_request_key(self) -> str:
        self.request_key += 1
        return f"{self.request_key}-{time.time_ns()}"



from omu.api.asset import (
    ASSET_PERMISSION_ID,
)
from omu.api.permission import PermissionType

ASSET_PERMISSION = PermissionType(
    id=ASSET_PERMISSION_ID,
    metadata={
        "level": "low",
        "name": {
            "ja": "ファイルを保持",
            "en": "Save a file",
        },
        "note": {
            "ja": "アプリがファイルを保持するために使われます",
            "en": "Used by apps to store files",
        },
    },
)



from __future__ import annotations

import asyncio
import json
import os
import subprocess
import sys
import time
from datetime import datetime
from urllib.parse import urlencode

import aiohttp
from aiohttp import web
from loguru import logger
from omu import Identifier
from omu.event_emitter import EventEmitter
from omu.helper import asyncio_error_logger
from omu.network.packet.packet_types import DisconnectType
from omu.result import Err, Ok
from yarl import URL

from omuserver.api.asset import AssetExtension
from omuserver.api.dashboard import DashboardExtension
from omuserver.api.endpoint import EndpointExtension
from omuserver.api.http import HttpExtension
from omuserver.api.i18n import I18nExtension
from omuserver.api.logger import LoggerExtension
from omuserver.api.permission import PermissionExtension
from omuserver.api.plugin import PluginExtension
from omuserver.api.registry import RegistryExtension
from omuserver.api.server import ServerExtension
from omuserver.api.session import SessionExtension
from omuserver.api.signal import SignalExtension
from omuserver.api.table import TableExtension
from omuserver.config import Config
from omuserver.consts import USER_AGENT_HEADERS
from omuserver.dependency import AppIndexRegistry, InstallRequest
from omuserver.network import Network
from omuserver.network.packet_dispatcher import ServerPacketDispatcher
from omuserver.security import PermissionManager
from omuserver.version import VERSION

RESTART_EXIT_CODE = 100
FRAME_TYPE_KEY = "omuapps-frame"


class Server:
    def __init__(
        self,
        config: Config,
        loop: asyncio.AbstractEventLoop | None = None,
    ) -> None:
        self.config = config
        self._loop = self._set_loop(loop or asyncio.new_event_loop())
        self.address = config.address
        self.event = ServerEvents()
        self.directories = config.directories
        self.directories.mkdir()
        self.packets = ServerPacketDispatcher()
        self.network = Network(self, self.packets)
        self.network.route_get("/", self._handle_index)
        self.network.route_get("/version", self._handle_version)
        self.network.route_get("/frame", self._handle_frame)
        self.network.route_get("/index/install", self._handle_get_index_install)
        self.network.route_post("/api/index/install", self._handle_post_index_install)
        self.security = PermissionManager.load(self)
        self.running = False
        self.endpoints = EndpointExtension(self)
        self.permissions = PermissionExtension(self)
        self.tables = TableExtension(self)
        self.sessions = SessionExtension(self)
        self.registries = RegistryExtension(self)
        self.dashboard = DashboardExtension(self)
        self.server = ServerExtension(self)
        self.http = HttpExtension(self)
        self.signals = SignalExtension(self)
        self.plugins = PluginExtension(self)
        self.assets = AssetExtension(self)
        self.i18n = I18nExtension(self)
        self.logger = LoggerExtension(self)
        self.client = aiohttp.ClientSession(
            loop=self.loop,
            headers=USER_AGENT_HEADERS,
            timeout=aiohttp.ClientTimeout(total=10),
        )
        self.last_prompt_timestamp: dict[str, float] = {}

    def _set_loop(self, loop: asyncio.AbstractEventLoop) -> asyncio.AbstractEventLoop:
        loop = asyncio.new_event_loop()
        loop.set_exception_handler(asyncio_error_logger)
        return loop

    async def _handle_index(self, request: web.Request) -> web.StreamResponse:
        return web.FileResponse(self.directories.index)

    async def _handle_version(self, request: web.Request) -> web.Response:
        return web.json_response({"version": VERSION})

    async def _handle_frame(self, request: web.Request) -> web.StreamResponse:
        url = request.query.get("url")
        if not url:
            return web.Response(status=400)
        url = URL(url).human_repr()
        content = self.directories.frame.read_text(encoding="utf-8")
        frame_token = self.security.generate_frame_token(url)
        config = {
            "frame_token": frame_token,
            "url": url,
            "ws_url": URL.build(
                scheme="ws",
                host=self.address.host,
                port=self.address.port,
                path="/ws",
                query_string=urlencode({"frame_token": frame_token, "url": url}),
            ).human_repr(),
            "type_key": FRAME_TYPE_KEY,
        }
        content = content.replace("%CONFIG%", json.dumps(config))
        return web.Response(text=content, content_type="text/html")

    async def _handle_get_index_install(self, request: web.Request) -> web.StreamResponse:
        # Check host
        if request.remote:
            if request.remote not in {"localhost", "127.0.0.1", self.address.host}:
                return self._error_response(f"Invalid Host: {request.remote} is not a local IP")

        raw_index_url = request.query.get("index_url")
        if not raw_index_url:
            return self._error_response("inedx_url not provided")
        raw_index_url = URL(raw_index_url)
        match await AppIndexRegistry.try_fetch(raw_index_url):
            case Err(err):
                return self._error_response(f"Failed to fetch index: {err}")
            case Ok(index):
                ...
        mapped_index_url = self.network.get_mapped_url(raw_index_url)
        provided_index_namespace = Identifier.namespace_from_url(mapped_index_url)

        if index.id.namespace != provided_index_namespace:
            return self._error_response(f"Namespace mismatch: {index.id.namespace} != {provided_index_namespace}")

        if not index.id.is_namepath_equal(index.id):
            return self._error_response("Trust issue")

        content = self.directories.index_install.read_text(encoding="utf-8")
        config = {
            "id": index.id.key(),
            "meta": index.meta,
            "url": str(raw_index_url),
        }
        server_index = self.server.index.get()
        already_installed = index.id.key() in server_index["indexes"]
        if already_installed:
            config["installed"] = True
        content = content.replace("%CONFIG%", json.dumps(config))
        return web.Response(text=content, content_type="text/html")

    def _error_response(self, message: str, status=400) -> web.StreamResponse:
        return web.json_response(
            {"type": "error", "message": message},
            status=status,
            reason=message,
        )

    async def _handle_post_index_install(self, request: web.Request) -> web.StreamResponse:
        # Check origin
        origin = request.headers.get("Origin")
        if origin is None:
            return self._error_response("Missing Origin header")
        origin_url = URL(origin)
        if origin_url.host not in {"localhost", "127.0.0.1", self.address.host}:
            return self._error_response("Invalid Origin Host")
        if origin_url.port not in {self.address.port}:
            return self._error_response("Invalid Origin Port")
        # Check host
        if request.remote:
            if request.remote not in {"localhost", "127.0.0.1", self.address.host}:
                return self._error_response(f"Invalid Host: {request.remote} is not a local IP")

        install_request: InstallRequest = await request.json()
        raw_index_url = URL(install_request["index"])
        key = Identifier.from_key(install_request["id"]).namespace
        elapsed = time.time() - self.last_prompt_timestamp.get(key, 0)
        remaining = 5 - elapsed
        if remaining > 0:
            return self._error_response(
                f"An installation is already in progress. You can try again in {remaining:.1f} seconds."
            )
        self.last_prompt_timestamp[key] = time.time()
        mapped_index_url = self.network.get_mapped_url(raw_index_url)
        provided_index_namespace = Identifier.namespace_from_url(mapped_index_url)

        match await AppIndexRegistry.try_fetch(raw_index_url):
            case Err(err):
                return self._error_response(f"Failed to fetch index: {err}")
            case Ok(index):
                ...

        if index.id.namespace != provided_index_namespace:
            return self._error_response(f"Namespace mismatch: {index.id.namespace} != {provided_index_namespace}")

        if not index.id.is_namepath_equal(index.id):
            return self._error_response("Trust issue")

        accepted = await self.dashboard.notify_index_install(mapped_index_url, index.meta)

        if not accepted:
            return self._error_response(
                "Installation denied by the user.",
                status=403,
            )

        server_index = self.server.index.get()
        server_index["indexes"][index.id.key()] = {
            "added_at": datetime.now().isoformat(),
            "meta": index.meta,
            "url": str(raw_index_url),
        }
        self.server.index.set(server_index)

        return web.json_response(
            {"type": "installed"},
            status=200,
        )

    def run(self) -> None:
        async def _run():
            await self.start()

        if self._loop is None:
            asyncio.run(_run())
        else:
            self._loop.create_task(_run())
            self._loop.run_forever()

    async def start(self) -> None:
        self.running = True
        try:
            await self.network.start()
            logger.info(f"Listening on {self.address.host}:{self.address.port}")
            await self.event.start()
        except Exception as e:
            logger.opt(exception=e).error("Failed to start server")
            await self.stop()
            self.loop.stop()
            raise e

    async def stop(self) -> None:
        logger.info("Stopping server")
        self.running = False
        await self.event.stop()
        await self.network.stop()

    async def restart(self) -> None:
        for session in list(self.sessions.iter()):
            if session.closed:
                continue
            await session.disconnect(DisconnectType.SERVER_RESTART, "Server is restarting")
        await self.stop()
        child = subprocess.Popen(
            args=[sys.executable, "-m", "omuserver", *sys.argv[1:]],
            cwd=os.getcwd(),
        )
        logger.info(f"Restarting server with PID {child.pid}")
        os._exit(RESTART_EXIT_CODE)

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop


class ServerEvents:
    def __init__(self) -> None:
        self.start = EventEmitter[[]]()
        self.stop = EventEmitter[[]]()



import datetime
import io
import os
import re
import sys
import threading
import time
import zipfile
from collections.abc import Generator
from pathlib import Path

import psutil
from loguru import logger
from omu.app import App
from omu.identifier import Identifier

LOG_DIRECTORY = Path("logs")


def _ensure_log_dir(path: Path) -> Path:
    path = path.resolve()
    if path.exists() and not path.is_dir():
        logger.error(f"Log directory {path} is not a directory")
        path.unlink()
    if not path.exists():
        path.mkdir(exist_ok=True, parents=True)
    return path


def _get_log_path(name: str, base_dir=LOG_DIRECTORY) -> Path:
    time_str = datetime.datetime.now().strftime("%H-%M-%S")
    date_str = datetime.datetime.now().strftime("%Y-%m-%d")
    base_dir = base_dir.resolve()
    path = base_dir / date_str / f"{time_str}-{name}.log"
    index = 1
    while path.exists():
        time_str = datetime.datetime.now().strftime("%H-%M-%S")
        path = base_dir / date_str / f"{time_str}-{name}-{index}.log"
        index += 1
    return path


def cleanup_logs(base_dir: Path) -> None:
    _cleanup_old_logs(base_dir)
    _remove_empty_folders(base_dir)


def _cleanup_old_logs(base_dir: Path) -> None:
    SEVEN_DAYS_IN_SECONDS = 60 * 60 * 24 * 7
    for file in base_dir.glob("**/*"):
        try:
            if not file.is_file():
                continue
            stats = file.stat()
            elapsed = time.time() - stats.st_mtime
            if elapsed > SEVEN_DAYS_IN_SECONDS:
                file.unlink()
        except (FileNotFoundError, IsADirectoryError):
            pass
        except PermissionError:
            logger.warning(f"Permission denied to delete log file {file}")
        except Exception as e:
            logger.opt(exception=e).error(f"Error deleting log file {file}")


def _remove_empty_folders(base_dir: Path) -> None:
    for file in base_dir.glob("**/*"):
        try:
            if not file.is_dir():
                continue
            if not any(file.iterdir()):
                file.rmdir()
        except (FileNotFoundError, IsADirectoryError):
            pass
        except PermissionError:
            logger.warning(f"Permission denied to delete log file {file}")
        except Exception as e:
            logger.opt(exception=e).error(f"Error deleting log file {file}")


COMPRESSING_THREAD: threading.Thread | None = None


def start_compressing_logs(base_dir: Path) -> None:
    global COMPRESSING_THREAD
    if COMPRESSING_THREAD and COMPRESSING_THREAD.is_alive():
        COMPRESSING_THREAD.join()
    thread = threading.Thread(target=compress_old_logs, args=(base_dir,))
    thread.daemon = True
    thread.start()
    COMPRESSING_THREAD = thread


def compress_old_logs(base_dir: Path) -> None:
    now = datetime.datetime.now()
    for folder in base_dir.glob("*"):
        try:
            if not folder.is_dir():
                continue
            date = datetime.datetime.strptime(folder.name, "%Y-%m-%d")
            elapsed = now - date
            if elapsed.days < 2:
                continue
            logs = list(folder.glob("*.log"))
            if not logs:
                continue
            dest_archive = folder.with_suffix(".zip")
            if dest_archive.exists():
                logger.warning(f"Log archive {dest_archive} already exists")
            with zipfile.ZipFile(dest_archive, mode="w", compression=zipfile.ZIP_DEFLATED, compresslevel=9) as f_out:
                for log_file in logs:
                    relative_path = log_file.relative_to(folder)
                    f_out.write(log_file, arcname=relative_path)
            for log_file in logs:
                log_file.unlink()
            folder.rmdir()
        except ValueError:
            pass
        except (FileNotFoundError, IsADirectoryError):
            pass
        except PermissionError:
            logger.warning(f"Permission denied to delete log file {folder}")
        except Exception as e:
            logger.opt(exception=e).error(f"Error deleting log file {folder}")


def setup_logger(name: str | App | Identifier, base_dir=LOG_DIRECTORY) -> Path:
    if isinstance(sys.stdout, io.TextIOWrapper):
        sys.stdout.reconfigure(encoding="utf-8")
    if isinstance(sys.stderr, io.TextIOWrapper):
        sys.stderr.reconfigure(encoding="utf-8")
    # <logdir>/<date>/<logger-name>-<time>.log
    if isinstance(name, App):
        name = name.id.get_sanitized_key()
    elif isinstance(name, Identifier):
        name = name.get_sanitized_key()
    _ensure_log_dir(base_dir)
    log_file_path = _get_log_path(name, base_dir=base_dir)
    logger.add(
        log_file_path,
        colorize=False,
        format=("{time:YYYY-MM-DD HH:mm:ss} | {level: <8} | " "{name}:{function}:{line} - {message}"),
        retention="7 days",
    )
    logger.info(f"Logging to {log_file_path}")
    cleanup_logs(base_dir)
    return log_file_path


def safe_path(root_path: Path, input_path: Path) -> Path:
    """
    How to prevent directory traversal attack from Python code
    https://stackoverflow.com/a/45190125
    """
    resolved_path = root_path.joinpath(input_path).resolve()
    if not resolved_path.is_relative_to(root_path.resolve()):
        raise ValueError(f"Path {input_path} is not relative to {root_path}")
    return resolved_path.relative_to(root_path.resolve())


def safe_path_join(root: Path, *paths: Path | str) -> Path:
    return root / safe_path(root, root.joinpath(*paths))


def find_processes_by_port(port: int) -> Generator[psutil.Process, None, None]:
    for connection in psutil.net_connections():
        try:
            if connection.laddr and connection.laddr.port == port:
                yield psutil.Process(connection.pid)
        except psutil.NoSuchProcess:
            pass
        except psutil.AccessDenied:
            pass


def find_processes_by_executable(
    executable: Path,
) -> Generator[psutil.Process, None, None]:
    for process in psutil.process_iter():
        try:
            if Path(process.exe()).resolve() == executable.resolve():
                yield process
        except psutil.NoSuchProcess:
            pass
        except psutil.AccessDenied:
            pass


def get_parent_pids() -> Generator[int]:
    pid = os.getpid()
    yield pid
    for parent in psutil.Process(pid).parents():
        yield parent.pid


def normalize_package_name(pkg: str) -> str:
    # https://packaging.python.org/en/latest/specifications/name-normalization/#name-normalization
    return re.sub(r"[-_.]+", "-", pkg).lower()



from omu.api.permission.permission import PermissionType
from omu.api.plugin import (
    PLUGIN_MANAGE_PACKAGE_PERMISSION_ID,
    PLUGIN_READ_PACKAGE_PERMISSION_ID,
)

PLUGIN_MANAGE_PERMISSION = PermissionType(
    PLUGIN_MANAGE_PACKAGE_PERMISSION_ID,
    metadata={
        "level": "high",
        "name": {
            "ja": "プラグインを管理",
            "en": "Manage packages",
        },
        "note": {
            "ja": "任意のプラグインの追加や更新、再読み込みを行う",
            "en": "Add, update, or reload any plugin",
        },
    },
)

PLUGIN_READ_PERMISSION = PermissionType(
    PLUGIN_READ_PACKAGE_PERMISSION_ID,
    metadata={
        "level": "low",
        "name": {
            "ja": "プラグインを取得",
            "en": "Read packages",
        },
        "note": {
            "ja": "インストールされているプラグインに関する情報を取得",
            "en": "Get information about installed plugins",
        },
    },
)
